# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: mjc_message.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'mjc_message.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11mjc_message.proto\x12\rMujocoMessage\"J\n\x19SetSimulationStateRequest\x12-\n\x05state\x18\x01 \x01(\x0e\x32\x1e.MujocoMessage.SimulationState\"\x1c\n\x1aSetSimulationStateResponse\"\x17\n\x15QueryModelInfoRequest\"\xb7\x02\n\x16QueryModelInfoResponse\x12\n\n\x02nq\x18\x01 \x01(\x05\x12\n\n\x02nv\x18\x02 \x01(\x05\x12\n\n\x02nu\x18\x03 \x01(\x05\x12\r\n\x05nbody\x18\x04 \x01(\x05\x12\x0c\n\x04njnt\x18\x05 \x01(\x05\x12\r\n\x05ngeom\x18\x06 \x01(\x05\x12\r\n\x05nsite\x18\x07 \x01(\x05\x12\r\n\x05nmesh\x18\x08 \x01(\x05\x12\x0c\n\x04ncam\x18\t \x01(\x05\x12\x0e\n\x06nlight\x18\n \x01(\x05\x12\x12\n\nnuser_body\x18\x0b \x01(\x05\x12\x11\n\tnuser_jnt\x18\x0c \x01(\x05\x12\x12\n\nnuser_geom\x18\r \x01(\x05\x12\x12\n\nnuser_site\x18\x0e \x01(\x05\x12\x14\n\x0cnuser_tendon\x18\x0f \x01(\x05\x12\x16\n\x0enuser_actuator\x18\x10 \x01(\x05\x12\x14\n\x0cnuser_sensor\x18\x11 \x01(\x05\"\x17\n\x15QueryOptConfigRequest\"\xdc\x04\n\x16QueryOptConfigResponse\x12\x10\n\x08timestep\x18\x01 \x01(\x01\x12\x0f\n\x07\x61pirate\x18\x02 \x01(\x01\x12\x10\n\x08impratio\x18\x03 \x01(\x01\x12\x11\n\ttolerance\x18\x04 \x01(\x01\x12\x14\n\x0cls_tolerance\x18\x05 \x01(\x01\x12\x18\n\x10noslip_tolerance\x18\x06 \x01(\x01\x12\x15\n\rccd_tolerance\x18\x07 \x01(\x01\x12\x0f\n\x07gravity\x18\x08 \x03(\x01\x12\x0c\n\x04wind\x18\t \x03(\x01\x12\x10\n\x08magnetic\x18\n \x03(\x01\x12\x0f\n\x07\x64\x65nsity\x18\x0b \x01(\x01\x12\x11\n\tviscosity\x18\x0c \x01(\x01\x12\x10\n\x08o_margin\x18\r \x01(\x01\x12\x10\n\x08o_solref\x18\x0e \x03(\x01\x12\x10\n\x08o_solimp\x18\x0f \x03(\x01\x12\x12\n\no_friction\x18\x10 \x03(\x01\x12\x12\n\nintegrator\x18\x11 \x01(\x05\x12\x0c\n\x04\x63one\x18\x12 \x01(\x05\x12\x10\n\x08jacobian\x18\x13 \x01(\x05\x12\x0e\n\x06solver\x18\x14 \x01(\x05\x12\x12\n\niterations\x18\x15 \x01(\x05\x12\x15\n\rls_iterations\x18\x16 \x01(\x05\x12\x19\n\x11noslip_iterations\x18\x17 \x01(\x05\x12\x16\n\x0e\x63\x63\x64_iterations\x18\x18 \x01(\x05\x12\x14\n\x0c\x64isableflags\x18\x19 \x01(\x05\x12\x13\n\x0b\x65nableflags\x18\x1a \x01(\x05\x12\x17\n\x0f\x64isableactuator\x18\x1b \x01(\x05\x12\x16\n\x0esdf_initpoints\x18\x1c \x01(\x05\x12\x16\n\x0esdf_iterations\x18\x1d \x01(\x05\"\xd9\x04\n\x13SetOptConfigRequest\x12\x10\n\x08timestep\x18\x01 \x01(\x01\x12\x0f\n\x07\x61pirate\x18\x02 \x01(\x01\x12\x10\n\x08impratio\x18\x03 \x01(\x01\x12\x11\n\ttolerance\x18\x04 \x01(\x01\x12\x14\n\x0cls_tolerance\x18\x05 \x01(\x01\x12\x18\n\x10noslip_tolerance\x18\x06 \x01(\x01\x12\x15\n\rccd_tolerance\x18\x07 \x01(\x01\x12\x0f\n\x07gravity\x18\x08 \x03(\x01\x12\x0c\n\x04wind\x18\t \x03(\x01\x12\x10\n\x08magnetic\x18\n \x03(\x01\x12\x0f\n\x07\x64\x65nsity\x18\x0b \x01(\x01\x12\x11\n\tviscosity\x18\x0c \x01(\x01\x12\x10\n\x08o_margin\x18\r \x01(\x01\x12\x10\n\x08o_solref\x18\x0e \x03(\x01\x12\x10\n\x08o_solimp\x18\x0f \x03(\x01\x12\x12\n\no_friction\x18\x10 \x03(\x01\x12\x12\n\nintegrator\x18\x11 \x01(\x05\x12\x0c\n\x04\x63one\x18\x12 \x01(\x05\x12\x10\n\x08jacobian\x18\x13 \x01(\x05\x12\x0e\n\x06solver\x18\x14 \x01(\x05\x12\x12\n\niterations\x18\x15 \x01(\x05\x12\x15\n\rls_iterations\x18\x16 \x01(\x05\x12\x19\n\x11noslip_iterations\x18\x17 \x01(\x05\x12\x16\n\x0e\x63\x63\x64_iterations\x18\x18 \x01(\x05\x12\x14\n\x0c\x64isableflags\x18\x19 \x01(\x05\x12\x13\n\x0b\x65nableflags\x18\x1a \x01(\x05\x12\x17\n\x0f\x64isableactuator\x18\x1b \x01(\x05\x12\x16\n\x0esdf_initpoints\x18\x1c \x01(\x05\x12\x16\n\x0esdf_iterations\x18\x1d \x01(\x05\"\x16\n\x14SetOptConfigResponse\"\x17\n\x15QueryAllBodiesRequest\"\x8a\x04\n\x16QueryAllBodiesResponse\x12\x41\n\tbody_info\x18\x01 \x03(\x0b\x32..MujocoMessage.QueryAllBodiesResponse.BodyInfo\x1a\xac\x03\n\x08\x42odyInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x11\n\tparent_id\x18\x03 \x01(\x05\x12\x0f\n\x07root_id\x18\x04 \x01(\x05\x12\x0f\n\x07weld_id\x18\x05 \x01(\x05\x12\x10\n\x08mocap_id\x18\x06 \x01(\x05\x12\x0f\n\x07jnt_num\x18\x07 \x01(\x05\x12\x0f\n\x07jnt_adr\x18\x08 \x01(\x05\x12\x0f\n\x07\x64of_num\x18\t \x01(\x05\x12\x0f\n\x07\x64of_adr\x18\n \x01(\x05\x12\x0f\n\x07tree_id\x18\x0b \x01(\x05\x12\x10\n\x08geom_num\x18\x0c \x01(\x05\x12\x10\n\x08geom_adr\x18\r \x01(\x05\x12\x0e\n\x06simple\x18\x0e \x01(\x08\x12\x12\n\nsame_frame\x18\x0f \x01(\x08\x12\x0b\n\x03pos\x18\x10 \x03(\x01\x12\x0c\n\x04quat\x18\x11 \x03(\x01\x12\x0c\n\x04ipos\x18\x12 \x03(\x01\x12\r\n\x05iquat\x18\x13 \x03(\x01\x12\x0c\n\x04mass\x18\x14 \x01(\x01\x12\x14\n\x0csubtree_mass\x18\x15 \x01(\x01\x12\x0f\n\x07inertia\x18\x16 \x03(\x01\x12\x12\n\ninv_weight\x18\x17 \x03(\x01\x12\x11\n\tgrav_comp\x18\x18 \x01(\x01\x12\x0e\n\x06margin\x18\x19 \x01(\x01\"\x17\n\x15QueryAllJointsRequest\"\x8d\x03\n\x16QueryAllJointsResponse\x12\x43\n\njoint_info\x18\x01 \x03(\x0b\x32/.MujocoMessage.QueryAllJointsResponse.JointInfo\x1a\xad\x02\n\tJointInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x0f\n\x07\x62ody_id\x18\x03 \x01(\x05\x12\x0c\n\x04type\x18\x04 \x01(\x05\x12\r\n\x05range\x18\x05 \x03(\x01\x12\x16\n\x0eqpos_idx_start\x18\x06 \x01(\x05\x12\x16\n\x0eqvel_idx_start\x18\x07 \x01(\x05\x12\r\n\x05group\x18\x08 \x01(\x05\x12\x0f\n\x07limited\x18\t \x01(\x08\x12\x15\n\ractfrclimited\x18\n \x01(\x08\x12\x0e\n\x06solref\x18\x0b \x03(\x01\x12\x0e\n\x06solimp\x18\x0c \x03(\x01\x12\x0b\n\x03pos\x18\r \x03(\x01\x12\x0c\n\x04\x61xis\x18\x0e \x03(\x01\x12\x11\n\tstiffness\x18\x0f \x01(\x01\x12\x13\n\x0b\x61\x63tfrcrange\x18\x10 \x03(\x01\x12\x0e\n\x06margin\x18\x11 \x01(\x01\"\x18\n\x16QueryJointNamesRequest\"-\n\x17QueryJointNamesResponse\x12\x12\n\nJointNames\x18\x01 \x03(\t\"-\n\x17QueryJointDofadrRequest\x12\x12\n\nJointNames\x18\x01 \x03(\t\"0\n\x18QueryJointDofadrResponse\x12\x14\n\x0cJointDofadrs\x18\x01 \x03(\x05\"/\n\x15QueryChildBodyRequest\x12\x16\n\x0eParentBodyName\x18\x01 \x01(\t\"2\n\x16QueryChildBodyResponse\x12\x18\n\x10\x43hildrenBodyName\x18\x01 \x03(\t\",\n\x14QueryBodyQposRequest\x12\x14\n\x0c\x42odyNameList\x18\x01 \x03(\t\",\n\tJointQpos\x12\x11\n\tJointName\x18\x01 \x01(\t\x12\x0c\n\x04Qpos\x18\x02 \x03(\x01\"F\n\x08\x42odyQpos\x12\x10\n\x08\x42odyName\x18\x01 \x01(\t\x12(\n\x06Joints\x18\x02 \x03(\x0b\x32\x18.MujocoMessage.JointQpos\"F\n\x15QueryBodyQposResponse\x12-\n\x0c\x42odyQposList\x18\x01 \x03(\x0b\x32\x17.MujocoMessage.BodyQpos\",\n\x14QueryBodyQvelRequest\x12\x14\n\x0c\x42odyNameList\x18\x01 \x03(\t\",\n\tJointQvel\x12\x11\n\tJointName\x18\x01 \x01(\t\x12\x0c\n\x04Qvel\x18\x02 \x03(\x01\"F\n\x08\x42odyQvel\x12\x10\n\x08\x42odyName\x18\x01 \x01(\t\x12(\n\x06Joints\x18\x02 \x03(\x0b\x32\x18.MujocoMessage.JointQvel\"F\n\x15QueryBodyQvelResponse\x12-\n\x0c\x42odyQvelList\x18\x01 \x03(\x0b\x32\x17.MujocoMessage.BodyQvel\"\xa0\x01\n\x16SetControlInputRequest\x12G\n\x08\x43ontrols\x18\x01 \x03(\x0b\x32\x35.MujocoMessage.SetControlInputRequest.ActuatorControl\x1a=\n\x0f\x41\x63tuatorControl\x12\x14\n\x0c\x41\x63tuatorName\x18\x01 \x01(\t\x12\x14\n\x0c\x43ontrolInput\x18\x02 \x01(\x01\"\x9a\x01\n\x17SetControlInputResponse\x12\x46\n\x07Results\x18\x01 \x03(\x0b\x32\x35.MujocoMessage.SetControlInputResponse.ActuatorResult\x1a\x37\n\x0e\x41\x63tuatorResult\x12\x14\n\x0c\x41\x63tuatorName\x18\x01 \x01(\t\x12\x0f\n\x07Success\x18\x02 \x01(\x08\"\x19\n\x17LoadInitialFrameRequest\"\x1a\n\x18LoadInitialFrameResponse\"+\n\x13LoadKeyFrameRequest\x12\x14\n\x0cKeyFrameName\x18\x01 \x01(\t\"\'\n\x14LoadKeyFrameResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"\x1a\n\x18QueryAllActuatorsRequest\"\x9f\x06\n\x19QueryAllActuatorsResponse\x12O\n\x10\x41\x63tuatorDataList\x18\x01 \x03(\x0b\x32\x35.MujocoMessage.QueryAllActuatorsResponse.ActuatorData\x1a\xb0\x05\n\x0c\x41\x63tuatorData\x12\x14\n\x0c\x41\x63tuatorName\x18\x01 \x01(\t\x12\x11\n\tJointName\x18\x02 \x01(\t\x12\x11\n\tGearRatio\x18\x03 \x01(\x01\x12\x16\n\x0e\x61\x63tuator_trnid\x18\x04 \x03(\x05\x12\x1c\n\x14\x61\x63tuator_ctrllimited\x18\x05 \x01(\x08\x12\x1d\n\x15\x61\x63tuator_forcelimited\x18\x06 \x01(\x08\x12\x1b\n\x13\x61\x63tuator_actlimited\x18\x07 \x01(\x08\x12\x1a\n\x12\x61\x63tuator_ctrlrange\x18\x08 \x03(\x01\x12\x1b\n\x13\x61\x63tuator_forcerange\x18\t \x03(\x01\x12\x19\n\x11\x61\x63tuator_actrange\x18\n \x03(\x01\x12\x18\n\x10\x61\x63tuator_trntype\x18\x0b \x01(\x05\x12\x18\n\x10\x61\x63tuator_dyntype\x18\x0c \x01(\x05\x12\x19\n\x11\x61\x63tuator_gaintype\x18\r \x01(\x05\x12\x19\n\x11\x61\x63tuator_biastype\x18\x0e \x01(\x05\x12\x17\n\x0f\x61\x63tuator_actadr\x18\x0f \x01(\x05\x12\x17\n\x0f\x61\x63tuator_actnum\x18\x10 \x01(\x05\x12\x16\n\x0e\x61\x63tuator_group\x18\x11 \x01(\x05\x12\x17\n\x0f\x61\x63tuator_dynprm\x18\x12 \x03(\x01\x12\x18\n\x10\x61\x63tuator_gainprm\x18\x13 \x03(\x01\x12\x18\n\x10\x61\x63tuator_biasprm\x18\x14 \x03(\x01\x12\x19\n\x11\x61\x63tuator_actearly\x18\x15 \x01(\x08\x12\x15\n\ractuator_gear\x18\x16 \x03(\x01\x12\x1c\n\x14\x61\x63tuator_cranklength\x18\x17 \x01(\x01\x12\x15\n\ractuator_acc0\x18\x18 \x01(\x01\x12\x18\n\x10\x61\x63tuator_length0\x18\x19 \x01(\x01\x12\x1c\n\x14\x61\x63tuator_lengthrange\x18\x1a \x03(\x01\"\x16\n\x14QueryAllGeomsRequest\"\x9f\x04\n\x15QueryAllGeomsResponse\x12\x45\n\x0egeom_data_list\x18\x01 \x03(\x0b\x32-.MujocoMessage.QueryAllGeomsResponse.GeomData\x1a\xbe\x03\n\x08GeomData\x12\x11\n\tgeom_name\x18\x01 \x01(\t\x12\x11\n\tbody_name\x18\x02 \x01(\t\x12\x11\n\tgeom_type\x18\x03 \x01(\x05\x12\x14\n\x0cgeom_contype\x18\x04 \x01(\x05\x12\x18\n\x10geom_conaffinity\x18\x05 \x01(\x05\x12\x13\n\x0bgeom_condim\x18\x06 \x01(\x05\x12\x13\n\x0bgeom_solmix\x18\x07 \x01(\x01\x12\x13\n\x0bgeom_solref\x18\x08 \x03(\x01\x12\x13\n\x0bgeom_solimp\x18\t \x03(\x01\x12\x11\n\tgeom_size\x18\n \x03(\x01\x12\x15\n\rgeom_friction\x18\x0b \x03(\x01\x12\x13\n\x0bgeom_dataid\x18\x0c \x01(\x05\x12\x12\n\ngeom_matid\x18\r \x01(\x05\x12\x12\n\ngeom_group\x18\x0e \x01(\x05\x12\x15\n\rgeom_priority\x18\x0f \x01(\x05\x12\x13\n\x0bgeom_plugin\x18\x10 \x01(\x05\x12\x16\n\x0egeom_sameframe\x18\x11 \x01(\x08\x12\x10\n\x08geom_pos\x18\x12 \x03(\x01\x12\x11\n\tgeom_quat\x18\x13 \x03(\x01\x12\x13\n\x0bgeom_margin\x18\x14 \x01(\x01\x12\x10\n\x08geom_gap\x18\x15 \x01(\x01\".\n\x15QueryJointQposRequest\x12\x15\n\rJointNameList\x18\x01 \x03(\t\"I\n\x16QueryJointQposResponse\x12/\n\rJointQposList\x18\x01 \x03(\x0b\x32\x18.MujocoMessage.JointQpos\".\n\x15QueryJointQvelRequest\x12\x15\n\rJointNameList\x18\x01 \x03(\t\"I\n\x16QueryJointQvelResponse\x12/\n\rJointQvelList\x18\x01 \x03(\x0b\x32\x18.MujocoMessage.JointQvel\"J\n\x1f\x43\x61lculateStateDifferenceRequest\x12\x12\n\nJointNames\x18\x01 \x03(\t\x12\x13\n\x0bInitialQpos\x18\x02 \x03(\x01\">\n\x0fStateDifference\x12\x11\n\tJointName\x18\x01 \x01(\t\x12\n\n\x02\x44q\x18\x02 \x03(\x01\x12\x0c\n\x04Qvel\x18\x03 \x03(\x01\"Y\n CalculateStateDifferenceResponse\x12\x35\n\rStateDiffList\x18\x01 \x03(\x0b\x32\x1e.MujocoMessage.StateDifference\"<\n\x17\x44ifferentiatePosRequest\x12\x13\n\x0bInitialQpos\x18\x01 \x03(\x01\x12\x0c\n\x04Qpos\x18\x02 \x03(\x01\"&\n\x18\x44ifferentiatePosResponse\x12\n\n\x02\x44q\x18\x01 \x03(\x01\";\n\x13TransitionFDRequest\x12\x0f\n\x07\x45psilon\x18\x01 \x01(\x01\x12\x13\n\x0b\x46lgCentered\x18\x02 \x01(\x08\"D\n\x14TransitionFDResponse\x12\n\n\x02nv\x18\x01 \x01(\x05\x12\n\n\x02nu\x18\x02 \x01(\x05\x12\t\n\x01\x41\x18\x03 \x03(\x01\x12\t\n\x01\x42\x18\x04 \x03(\x01\"(\n\x14JacSubtreeComRequest\x12\x10\n\x08\x42odyName\x18\x01 \x01(\t\"\'\n\x15JacSubtreeComResponse\x12\x0e\n\x06JacCom\x18\x01 \x03(\x01\"%\n\x11JacBodyComRequest\x12\x10\n\x08\x42odyName\x18\x01 \x01(\t\"0\n\x12JacBodyComResponse\x12\x0c\n\x04JacP\x18\x01 \x03(\x01\x12\x0c\n\x04JacR\x18\x02 \x03(\x01\"\x1d\n\x1bQueryAllQposQvelQaccRequest\"H\n\x1cQueryAllQposQvelQaccResponse\x12\x0c\n\x04qpos\x18\x01 \x03(\x01\x12\x0c\n\x04qvel\x18\x02 \x03(\x01\x12\x0c\n\x04qacc\x18\x03 \x03(\x01\"\x15\n\x13QueryContactRequest\"\x94\x03\n\x14QueryContactResponse\x12\x41\n\x08\x63ontacts\x18\x01 \x03(\x0b\x32/.MujocoMessage.QueryContactResponse.ContactData\x1a\xb8\x02\n\x0b\x43ontactData\x12\x0c\n\x04\x64ist\x18\x01 \x01(\x01\x12\x0b\n\x03pos\x18\x02 \x03(\x01\x12\r\n\x05\x66rame\x18\x03 \x03(\x01\x12\x15\n\rincludemargin\x18\x04 \x01(\x01\x12\x10\n\x08\x66riction\x18\x05 \x03(\x01\x12\x0e\n\x06solref\x18\x06 \x03(\x01\x12\x16\n\x0esolreffriction\x18\x07 \x03(\x01\x12\x0e\n\x06solimp\x18\x08 \x03(\x01\x12\n\n\x02mu\x18\t \x01(\x01\x12\t\n\x01H\x18\n \x03(\x01\x12\x0b\n\x03\x64im\x18\x0b \x01(\x05\x12\r\n\x05geom1\x18\x0c \x01(\x05\x12\r\n\x05geom2\x18\r \x01(\x05\x12\x0c\n\x04geom\x18\x0e \x03(\x05\x12\x0c\n\x04\x66lex\x18\x0f \x03(\x05\x12\x0c\n\x04\x65lem\x18\x10 \x03(\x05\x12\x0c\n\x04vert\x18\x11 \x03(\x05\x12\x0f\n\x07\x65xclude\x18\x12 \x01(\x05\x12\x13\n\x0b\x65\x66\x63_address\x18\x13 \x01(\x05\"\x1b\n\x19QueryContactSimpleRequest\"\x9f\x01\n\x1aQueryContactSimpleResponse\x12G\n\x08\x63ontacts\x18\x01 \x03(\x0b\x32\x35.MujocoMessage.QueryContactSimpleResponse.ContactData\x1a\x38\n\x0b\x43ontactData\x12\x0b\n\x03\x64im\x18\x01 \x01(\x05\x12\r\n\x05geom1\x18\x02 \x01(\x05\x12\r\n\x05geom2\x18\x03 \x01(\x05\"\x19\n\x17QueryQfrcInverseRequest\"0\n\x18QueryQfrcInverseResponse\x12\x14\n\x0cqfrc_inverse\x18\x01 \x03(\x01\"\x1c\n\x1aQueryActuatorMomentRequest\"N\n\x1bQueryActuatorMomentResponse\x12\x17\n\x0f\x61\x63tuator_moment\x18\x01 \x03(\x01\x12\n\n\x02nv\x18\x02 \x01(\x05\x12\n\n\x02nu\x18\x03 \x01(\x05\"\x1a\n\x18QueryQfrcActuatorRequest\"2\n\x19QueryQfrcActuatorResponse\x12\x15\n\rqfrc_actuator\x18\x01 \x03(\x01\"\x1e\n\x0eSetQposRequest\x12\x0c\n\x04qpos\x18\x01 \x03(\x01\"\x11\n\x0fSetQposResponse\"\x1e\n\x0eSetQaccRequest\x12\x0c\n\x04qacc\x18\x01 \x03(\x01\"\x11\n\x0fSetQaccResponse\"6\n!QueryBodySubtreeMassByNameRequest\x12\x11\n\tbody_name\x18\x01 \x01(\t\">\n\"QueryBodySubtreeMassByNameResponse\x12\x18\n\x10\x62ody_subtreemass\x18\x01 \x01(\x01\")\n\x15SetOptTimestepRequest\x12\x10\n\x08timestep\x18\x01 \x01(\x01\"\x18\n\x16SetOptTimestepResponse\"\x1e\n\x0eSetCtrlRequest\x12\x0c\n\x04\x63trl\x18\x01 \x03(\x01\"\x11\n\x0fSetCtrlResponse\"\x13\n\x11MJ_ForwardRequest\"\x14\n\x12MJ_ForwardResponse\"\x13\n\x11MJ_InverseRequest\"\x14\n\x12MJ_InverseResponse\"\x1f\n\x0eMJ_StepRequest\x12\r\n\x05nstep\x18\x01 \x01(\x05\"\x11\n\x0fMJ_StepResponse\"\x1e\n\x0eSetQvelRequest\x12\x0c\n\x04qvel\x18\x01 \x03(\x01\"\x11\n\x0fSetQvelResponse\")\n\x13QueryCfrcExtRequest\x12\x12\n\nbody_names\x18\x01 \x03(\t\"[\n\x14QueryCfrcExtResponse\x12\x32\n\x0e\x62ody_cfrc_exts\x18\x01 \x03(\x0b\x32\x1a.MujocoMessage.BodyCfrcExt\x12\x0f\n\x07success\x18\x02 \x01(\x08\"2\n\x0b\x42odyCfrcExt\x12\x11\n\tbody_name\x18\x01 \x01(\t\x12\x10\n\x08\x63\x66rc_ext\x18\x02 \x03(\x01\"H\n\x13SetJointQposRequest\x12\x31\n\x0fjoint_qpos_list\x18\x01 \x03(\x0b\x32\x18.MujocoMessage.JointQpos\"\'\n\x14SetJointQposResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"\x1b\n\x19QueryActuatorForceRequest\"4\n\x1aQueryActuatorForceResponse\x12\x16\n\x0e\x61\x63tuator_force\x18\x01 \x03(\x01\".\n\x17QueryJointLimitsRequest\x12\x13\n\x0bjoint_names\x18\x01 \x03(\t\"Y\n\nJointLimit\x12\x12\n\njoint_name\x18\x01 \x01(\t\x12\x11\n\thas_limit\x18\x02 \x01(\x08\x12\x11\n\trange_min\x18\x03 \x01(\x01\x12\x11\n\trange_max\x18\x04 \x01(\x01\"K\n\x18QueryJointLimitsResponse\x12/\n\x0cjoint_limits\x18\x01 \x03(\x0b\x32\x19.MujocoMessage.JointLimit\"0\n\x1aQueryBodyVelocitiesRequest\x12\x12\n\nbody_names\x18\x01 \x03(\t\"\xc5\x01\n\x1bQueryBodyVelocitiesResponse\x12P\n\x0f\x62ody_velocities\x18\x01 \x03(\x0b\x32\x37.MujocoMessage.QueryBodyVelocitiesResponse.BodyVelocity\x1aT\n\x0c\x42odyVelocity\x12\x11\n\tbody_name\x18\x01 \x01(\t\x12\x17\n\x0flinear_velocity\x18\x02 \x03(\x01\x12\x18\n\x10\x61ngular_velocity\x18\x03 \x03(\x01\"5\n\x1bQueryActuatorGainPrmRequest\x12\x16\n\x0e\x61\x63tuator_names\x18\x01 \x03(\t\":\n\x0f\x41\x63tuatorGainPrm\x12\x15\n\ractuator_name\x18\x01 \x01(\t\x12\x10\n\x08gain_prm\x18\x02 \x03(\x01\"U\n\x1cQueryActuatorGainPrmResponse\x12\x35\n\rgain_prm_list\x18\x01 \x03(\x0b\x32\x1e.MujocoMessage.ActuatorGainPrm\"\xb2\x01\n\x19SetActuatorGainPrmRequest\x12V\n\x11gain_prm_set_list\x18\x01 \x03(\x0b\x32;.MujocoMessage.SetActuatorGainPrmRequest.ActuatorGainPrmSet\x1a=\n\x12\x41\x63tuatorGainPrmSet\x12\x15\n\ractuator_name\x18\x01 \x01(\t\x12\x10\n\x08gain_prm\x18\x02 \x03(\x01\"-\n\x1aSetActuatorGainPrmResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"5\n\x1bQueryActuatorBiasPrmRequest\x12\x16\n\x0e\x61\x63tuator_names\x18\x01 \x03(\t\":\n\x0f\x41\x63tuatorBiasPrm\x12\x15\n\ractuator_name\x18\x01 \x01(\t\x12\x10\n\x08\x62ias_prm\x18\x02 \x03(\x01\"U\n\x1cQueryActuatorBiasPrmResponse\x12\x35\n\rbias_prm_list\x18\x01 \x03(\x0b\x32\x1e.MujocoMessage.ActuatorBiasPrm\"\xb2\x01\n\x19SetActuatorBiasPrmRequest\x12V\n\x11\x62ias_prm_set_list\x18\x01 \x03(\x0b\x32;.MujocoMessage.SetActuatorBiasPrmRequest.ActuatorBiasPrmSet\x1a=\n\x12\x41\x63tuatorBiasPrmSet\x12\x15\n\ractuator_name\x18\x01 \x01(\t\x12\x10\n\x08\x62ias_prm\x18\x02 \x03(\x01\"-\n\x1aSetActuatorBiasPrmResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"\x1c\n\x1aQueryAllMocapBodiesRequest\"\xae\x01\n\x1bQueryAllMocapBodiesResponse\x12N\n\x0cmocap_bodies\x18\x01 \x03(\x0b\x32\x38.MujocoMessage.QueryAllMocapBodiesResponse.MocapBodyInfo\x1a?\n\rMocapBodyInfo\x12\x17\n\x0fmocap_body_name\x18\x01 \x01(\t\x12\x15\n\rmocap_body_id\x18\x02 \x01(\x05\"7\n\x1bQueryMocapPosAndQuatRequest\x12\x18\n\x10mocap_body_names\x18\x01 \x03(\t\"\xb7\x01\n\x1cQueryMocapPosAndQuatResponse\x12R\n\x0fmocap_body_info\x18\x01 \x03(\x0b\x32\x39.MujocoMessage.QueryMocapPosAndQuatResponse.MocapBodyInfo\x1a\x43\n\rMocapBodyInfo\x12\x17\n\x0fmocap_body_name\x18\x01 \x01(\t\x12\x0b\n\x03pos\x18\x02 \x03(\x01\x12\x0c\n\x04quat\x18\x03 \x03(\x01\"\xb1\x01\n\x19SetMocapPosAndQuatRequest\x12O\n\x0fmocap_body_info\x18\x01 \x03(\x0b\x32\x36.MujocoMessage.SetMocapPosAndQuatRequest.MocapBodyInfo\x1a\x43\n\rMocapBodyInfo\x12\x17\n\x0fmocap_body_name\x18\x01 \x01(\t\x12\x0b\n\x03pos\x18\x02 \x03(\x01\x12\x0c\n\x04quat\x18\x03 \x03(\x01\"-\n\x1aSetMocapPosAndQuatResponse\x12\x0f\n\x07success\x18\x01 \x03(\x08\"$\n\"QueryAllEqualityConstraintsRequest\"\x9a\x02\n#QueryAllEqualityConstraintsResponse\x12g\n\x14\x65quality_constraints\x18\x01 \x03(\x0b\x32I.MujocoMessage.QueryAllEqualityConstraintsResponse.EqualityConstraintInfo\x1a\x89\x01\n\x16\x45qualityConstraintInfo\x12\x0f\n\x07\x65q_type\x18\x01 \x01(\x05\x12\x0f\n\x07obj1_id\x18\x02 \x01(\x05\x12\x0f\n\x07obj2_id\x18\x03 \x01(\x05\x12\x0e\n\x06\x61\x63tive\x18\x04 \x01(\x08\x12\x0e\n\x06solref\x18\x05 \x03(\x01\x12\x0e\n\x06solimp\x18\x06 \x03(\x01\x12\x0c\n\x04\x64\x61ta\x18\x07 \x03(\x01\"/\n\x19QuerySitePosAndMatRequest\x12\x12\n\nsite_names\x18\x01 \x03(\t\"\xaf\x01\n\x1aQuerySitePosAndMatResponse\x12Q\n\x10site_pos_and_mat\x18\x01 \x03(\x0b\x32\x37.MujocoMessage.QuerySitePosAndMatResponse.SitePosAndMat\x1a>\n\rSitePosAndMat\x12\x11\n\tsite_name\x18\x01 \x01(\t\x12\x0c\n\x04xpos\x18\x02 \x03(\x01\x12\x0c\n\x04xmat\x18\x03 \x03(\x01\")\n\x13QuerySiteJacRequest\x12\x12\n\nsite_names\x18\x01 \x03(\t\"\x90\x01\n\x14QuerySiteJacResponse\x12>\n\tsite_jacs\x18\x01 \x03(\x0b\x32+.MujocoMessage.QuerySiteJacResponse.SiteJac\x1a\x38\n\x07SiteJac\x12\x11\n\tsite_name\x18\x01 \x01(\t\x12\x0c\n\x04jacp\x18\x02 \x03(\x01\x12\x0c\n\x04jacr\x18\x03 \x03(\x01\"\xd4\x01\n UpdateEqualityConstraintsRequest\x12\x65\n\x14\x65quality_constraints\x18\x01 \x03(\x0b\x32G.MujocoMessage.UpdateEqualityConstraintsRequest.EqualityConstraintsInfo\x1aI\n\x17\x45qualityConstraintsInfo\x12\x0f\n\x07obj1_id\x18\x01 \x01(\x05\x12\x0f\n\x07obj2_id\x18\x02 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x03 \x03(\x01\"4\n!UpdateEqualityConstraintsResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"3\n\x19QueryBodyComPosMatRequest\x12\x16\n\x0e\x62ody_name_list\x18\x01 \x03(\t\"\xba\x01\n\x1aQueryBodyComPosMatResponse\x12V\n\x15\x62ody_com_pos_mat_list\x18\x01 \x03(\x0b\x32\x37.MujocoMessage.QueryBodyComPosMatResponse.BodyComPosMat\x1a\x44\n\rBodyComPosMat\x12\x11\n\tbody_name\x18\x01 \x01(\t\x12\x0f\n\x07\x63om_pos\x18\x02 \x03(\x01\x12\x0f\n\x07\x63om_mat\x18\x03 \x03(\x01\"4\n\x1aQueryBodyPosMatQuatRequest\x12\x16\n\x0e\x62ody_name_list\x18\x01 \x03(\t\"\xc5\x01\n\x1bQueryBodyPosMatQuatResponse\x12Y\n\x16\x62ody_pos_mat_quat_list\x18\x01 \x03(\x0b\x32\x39.MujocoMessage.QueryBodyPosMatQuatResponse.BodyPosMatQuat\x1aK\n\x0e\x42odyPosMatQuat\x12\x11\n\tbody_name\x18\x01 \x01(\t\x12\x0b\n\x03pos\x18\x02 \x03(\x01\x12\x0b\n\x03mat\x18\x03 \x03(\x01\x12\x0c\n\x04quat\x18\x04 \x03(\x01\"0\n\x16QueryGeomPosMatRequest\x12\x16\n\x0egeom_name_list\x18\x01 \x03(\t\"\xa2\x01\n\x17QueryGeomPosMatResponse\x12L\n\x11geom_pos_mat_list\x18\x01 \x03(\x0b\x32\x31.MujocoMessage.QueryGeomPosMatResponse.GeomPosMat\x1a\x39\n\nGeomPosMat\x12\x11\n\tgeom_name\x18\x01 \x01(\t\x12\x0b\n\x03pos\x18\x02 \x03(\x01\x12\x0b\n\x03mat\x18\x03 \x03(\x01\"/\n\x18QueryContactForceRequest\x12\x13\n\x0b\x63ontact_ids\x18\x01 \x03(\x05\"\x96\x01\n\x19QueryContactForceResponse\x12M\n\x0e\x63ontact_forces\x18\x01 \x03(\x0b\x32\x35.MujocoMessage.QueryContactForceResponse.ContactForce\x1a*\n\x0c\x43ontactForce\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0e\n\x06\x66orces\x18\x02 \x03(\x01\"\xa6\x01\n\rMJ_JacRequest\x12?\n\x0f\x62ody_point_list\x18\x01 \x03(\x0b\x32&.MujocoMessage.MJ_JacRequest.BodyPoint\x12\x14\n\x0c\x63ompute_jacp\x18\x02 \x01(\x08\x12\x14\n\x0c\x63ompute_jacr\x18\x03 \x01(\x08\x1a(\n\tBodyPoint\x12\r\n\x05point\x18\x01 \x03(\x01\x12\x0c\n\x04\x62ody\x18\x02 \x01(\x05\"w\n\x0eMJ_JacResponse\x12<\n\x0bjac_results\x18\x01 \x03(\x0b\x32\'.MujocoMessage.MJ_JacResponse.JacResult\x1a\'\n\tJacResult\x12\x0c\n\x04jacp\x18\x01 \x03(\x01\x12\x0c\n\x04jacr\x18\x02 \x03(\x01\"\x1b\n\x19\x43\x61lcFullMassMatrixRequest\"6\n\x1a\x43\x61lcFullMassMatrixResponse\x12\x18\n\x10\x66ull_mass_matrix\x18\x01 \x03(\x01\"\x16\n\x14QueryQfrcBiasRequest\"*\n\x15QueryQfrcBiasResponse\x12\x11\n\tqfrc_bias\x18\x01 \x03(\x01\"0\n\x16QuerySubtreeComRequest\x12\x16\n\x0e\x62ody_name_list\x18\x01 \x03(\t\"F\n\x17QuerySubtreeComResponse\x12\x16\n\x0e\x62ody_name_list\x18\x01 \x03(\t\x12\x13\n\x0bsubtree_com\x18\x02 \x03(\x01\"\x93\x01\n\x16SetGeomFrictionRequest\x12\x16\n\x0egeom_name_list\x18\x01 \x03(\t\x12\x45\n\rfriction_list\x18\x02 \x03(\x0b\x32..MujocoMessage.SetGeomFrictionRequest.Friction\x1a\x1a\n\x08\x46riction\x12\x0e\n\x06values\x18\x01 \x03(\x01\"/\n\x17SetGeomFrictionResponse\x12\x14\n\x0csuccess_list\x18\x01 \x03(\x08\".\n\x16QuerySensorDataRequest\x12\x14\n\x0csensor_names\x18\x01 \x03(\t\"\xb4\x01\n\x17QuerySensorDataResponse\x12K\n\x10sensor_data_list\x18\x01 \x03(\x0b\x32\x31.MujocoMessage.QuerySensorDataResponse.SensorData\x1aL\n\nSensorData\x12\x13\n\x0bsensor_name\x18\x01 \x01(\t\x12\x13\n\x0bsensor_type\x18\x02 \x01(\t\x12\x14\n\x0csensor_value\x18\x03 \x03(\x01\"/\n\x18QueryJointOffsetsRequest\x12\x13\n\x0bjoint_names\x18\x01 \x03(\t\"\xca\x01\n\x19QueryJointOffsetsResponse\x12K\n\rjoint_offsets\x18\x01 \x03(\x0b\x32\x34.MujocoMessage.QueryJointOffsetsResponse.JointOffset\x1a`\n\x0bJointOffset\x12\x12\n\njoint_name\x18\x01 \x01(\t\x12\x13\n\x0bqpos_offset\x18\x02 \x01(\x05\x12\x13\n\x0bqvel_offset\x18\x03 \x01(\x05\x12\x13\n\x0bqacc_offset\x18\x04 \x01(\x05\"\x16\n\x14QueryAllSitesRequest\"\xe9\x01\n\x15QueryAllSitesResponse\x12@\n\tsite_info\x18\x01 \x03(\x0b\x32-.MujocoMessage.QueryAllSitesResponse.SiteInfo\x1a\x8d\x01\n\x08SiteInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x0b\n\x03pos\x18\x03 \x03(\x01\x12\x0b\n\x03mat\x18\x04 \x03(\x01\x12\x13\n\x0bsite_bodyid\x18\x05 \x01(\x05\x12\x11\n\tsite_type\x18\x06 \x01(\x05\x12\x11\n\tlocal_pos\x18\x07 \x03(\x01\x12\x12\n\nlocal_quat\x18\x08 \x03(\x01\"3\n\x15UpdateLocalEnvRequest\x12\x0c\n\x04qpos\x18\x01 \x03(\x01\x12\x0c\n\x04time\x18\x02 \x01(\x01\"\x9a\x01\n\x16UpdateLocalEnvResponse\x12N\n\x0eoverride_ctrls\x18\x01 \x03(\x0b\x32\x36.MujocoMessage.UpdateLocalEnvResponse.OverrideCtrlInfo\x1a\x30\n\x10OverrideCtrlInfo\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x01\"\x87\x01\n\x13LoadLocalEnvRequest\x12<\n\x08req_type\x18\x01 \x01(\x0e\x32*.MujocoMessage.LoadLocalEnvRequest.ReqType\"2\n\x07ReqType\x12\x11\n\rXML_FILE_NAME\x10\x00\x12\x14\n\x10XML_FILE_CONTENT\x10\x01\"\xb3\x01\n\x14LoadLocalEnvResponse\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x13\n\x0bxml_content\x18\x02 \x01(\x0c\x12:\n\x06status\x18\x03 \x01(\x0e\x32*.MujocoMessage.LoadLocalEnvResponse.Status\x12\x15\n\rerror_message\x18\x04 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"=\n\x16LoadContentFileRequest\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x10\n\x08\x66ile_dir\x18\x02 \x01(\t\"\xa2\x01\n\x17LoadContentFileResponse\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12=\n\x06status\x18\x02 \x01(\x0e\x32-.MujocoMessage.LoadContentFileResponse.Status\x12\x15\n\rerror_message\x18\x03 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"a\n\x0f\x41\x64\x64\x41\x63torRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x16\n\x0espawnable_name\x18\x02 \x01(\t\x12\x0b\n\x03pos\x18\x03 \x03(\x01\x12\x0c\n\x04quat\x18\x04 \x03(\x01\x12\r\n\x05scale\x18\x05 \x01(\x01\"\x83\x01\n\x10\x41\x64\x64\x41\x63torResponse\x12\x36\n\x06status\x18\x01 \x01(\x0e\x32&.MujocoMessage.AddActorResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"\x15\n\x13PublishSceneRequest\"\x8b\x01\n\x14PublishSceneResponse\x12:\n\x06status\x18\x01 \x01(\x0e\x32*.MujocoMessage.PublishSceneResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"W\n\x13SetLightInfoRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x13\n\x0blight_color\x18\x02 \x03(\x02\x12\x17\n\x0flight_intensity\x18\x03 \x01(\x02\"\x8b\x01\n\x14SetLightInfoResponse\x12:\n\x06status\x18\x01 \x01(\x0e\x32*.MujocoMessage.SetLightInfoResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"\x84\x01\n\x1aSetCameraSensorInfoRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x13\n\x0b\x63\x61pture_rgb\x18\x02 \x01(\x08\x12\x15\n\rcapture_depth\x18\x03 \x01(\x08\x12\x15\n\rsave_mp4_file\x18\x04 \x01(\x08\x12\x0f\n\x07use_dds\x18\x05 \x01(\x08\"\x99\x01\n\x1bSetCameraSensorInfoResponse\x12\x41\n\x06status\x18\x01 \x01(\x0e\x32\x31.MujocoMessage.SetCameraSensorInfoResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"\x9a\x01\n\x17\x42\x65ginSaveMp4FileRequest\x12\x11\n\tfile_path\x18\x01 \x01(\t\x12H\n\x0c\x63\x61pture_mode\x18\x02 \x01(\x0e\x32\x32.MujocoMessage.BeginSaveMp4FileRequest.CaptureMode\"\"\n\x0b\x43\x61ptureMode\x12\t\n\x05\x41SYNC\x10\x00\x12\x08\n\x04SYNC\x10\x01\"\x93\x01\n\x18\x42\x65ginSaveMp4FileResponse\x12>\n\x06status\x18\x01 \x01(\x0e\x32..MujocoMessage.BeginSaveMp4FileResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"y\n\x16StopSaveMp4FileRequest\x12<\n\x06status\x18\x01 \x01(\x0e\x32,.MujocoMessage.StopSaveMp4FileRequest.Status\"!\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\"\x19\n\x17StopSaveMp4FileResponse\"\x1d\n\x1bGetCurrentFrameIndexRequest\"L\n\x1cGetCurrentFrameIndexResponse\x12\x15\n\rcurrent_frame\x18\x01 \x01(\x03\x12\x15\n\rerror_message\x18\x02 \x01(\t\"/\n\x13GetTimeStampRequest\x12\x18\n\x10last_frame_index\x18\x01 \x01(\x04\"$\n\rTimeStampList\x12\x13\n\x0btime_stamps\x18\x01 \x03(\x04\"\xcf\x01\n\x14GetTimeStampResponse\x12M\n\x0etime_stamp_map\x18\x01 \x03(\x0b\x32\x35.MujocoMessage.GetTimeStampResponse.TimeStampMapEntry\x12\x15\n\rerror_message\x18\x02 \x01(\t\x1aQ\n\x11TimeStampMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12+\n\x05value\x18\x02 \x01(\x0b\x32\x1c.MujocoMessage.TimeStampList:\x02\x38\x01\".\n\x18GetCameraFramePNGRequest\x12\x12\n\nimage_path\x18\x01 \x01(\t\";\n\x10NameAndTransform\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03pos\x18\x02 \x03(\x01\x12\x0c\n\x04quat\x18\x03 \x03(\x01\"T\n\x19GetCameraFramePNGResponse\x12\x37\n\x0ename_transform\x18\x01 \x03(\x0b\x32\x1f.MujocoMessage.NameAndTransform\"J\n\x1fMakeCameraViewportActiveRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x13\n\x0b\x65ntity_name\x18\x02 \x01(\t\"\xa3\x01\n MakeCameraViewportActiveResponse\x12\x46\n\x06status\x18\x01 \x01(\x0e\x32\x36.MujocoMessage.MakeCameraViewportActiveResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"@\n\x16SetMaterialInfoRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x12\n\nbase_color\x18\x04 \x03(\x02\"\x91\x01\n\x17SetMaterialInfoResponse\x12=\n\x06status\x18\x01 \x01(\x0e\x32-.MujocoMessage.SetMaterialInfoResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"\x1c\n\x1aGetKeyPressedEventsRequest\"-\n\x1bGetKeyPressedEventsResponse\x12\x0e\n\x06\x65vents\x18\x01 \x03(\t\"\x1f\n\x1dGetMouseMovementEventsRequest\"S\n\x1eGetMouseMovementEventsResponse\x12\x0f\n\x07\x64\x65lta_x\x18\x01 \x01(\x02\x12\x0f\n\x07\x64\x65lta_y\x18\x02 \x01(\x02\x12\x0f\n\x07\x64\x65lta_z\x18\x03 \x01(\x02\"$\n\"GetBodyManipulationAnchoredRequest\"\xb8\x01\n#GetBodyManipulationAnchoredResponse\x12\x11\n\tbody_name\x18\x01 \x01(\t\x12R\n\x0b\x61nchor_type\x18\x02 \x01(\x0e\x32=.MujocoMessage.GetBodyManipulationAnchoredResponse.AnchorType\"*\n\nAnchorType\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04WELD\x10\x01\x12\x08\n\x04\x42\x41LL\x10\x02\"$\n\"GetBodyManipulationMovementRequest\"L\n#GetBodyManipulationMovementResponse\x12\x11\n\tdelta_pos\x18\x01 \x03(\x01\x12\x12\n\ndelta_quat\x18\x02 \x03(\x01\"W\n\x1eSetActorAnimParamNumberRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x12\n\nparam_name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x01\"\xa1\x01\n\x1fSetActorAnimParamNumberResponse\x12\x45\n\x06status\x18\x01 \x01(\x0e\x32\x35.MujocoMessage.SetActorAnimParamNumberResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"U\n\x1cSetActorAnimParamBoolRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x12\n\nparam_name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x08\"\x9d\x01\n\x1dSetActorAnimParamBoolResponse\x12\x43\n\x06status\x18\x01 \x01(\x0e\x32\x33.MujocoMessage.SetActorAnimParamBoolResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"W\n\x1eSetActorAnimParamStringRequest\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x12\n\nparam_name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"\xa1\x01\n\x1fSetActorAnimParamStringResponse\x12\x45\n\x06status\x18\x01 \x01(\x0e\x32\x35.MujocoMessage.SetActorAnimParamStringResponse.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\" \n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01**\n\x0fSimulationState\x12\x0b\n\x07RUNNING\x10\x00\x12\n\n\x06PAUSED\x10\x01\x32\xc1\x44\n\x0bGrpcService\x12i\n\x12SetSimulationState\x12(.MujocoMessage.SetSimulationStateRequest\x1a).MujocoMessage.SetSimulationStateResponse\x12\x63\n\x10LoadInitialFrame\x12&.MujocoMessage.LoadInitialFrameRequest\x1a\'.MujocoMessage.LoadInitialFrameResponse\x12W\n\x0cLoadKeyFrame\x12\".MujocoMessage.LoadKeyFrameRequest\x1a#.MujocoMessage.LoadKeyFrameResponse\x12]\n\x0eUpdateLocalEnv\x12$.MujocoMessage.UpdateLocalEnvRequest\x1a%.MujocoMessage.UpdateLocalEnvResponse\x12W\n\x0cLoadLocalEnv\x12\".MujocoMessage.LoadLocalEnvRequest\x1a#.MujocoMessage.LoadLocalEnvResponse\x12`\n\x0fLoadContentFile\x12%.MujocoMessage.LoadContentFileRequest\x1a&.MujocoMessage.LoadContentFileResponse\x12K\n\x08\x41\x64\x64\x41\x63tor\x12\x1e.MujocoMessage.AddActorRequest\x1a\x1f.MujocoMessage.AddActorResponse\x12W\n\x0cPublishScene\x12\".MujocoMessage.PublishSceneRequest\x1a#.MujocoMessage.PublishSceneResponse\x12W\n\x0cSetLightInfo\x12\".MujocoMessage.SetLightInfoRequest\x1a#.MujocoMessage.SetLightInfoResponse\x12l\n\x13SetCameraSensorInfo\x12).MujocoMessage.SetCameraSensorInfoRequest\x1a*.MujocoMessage.SetCameraSensorInfoResponse\x12{\n\x18MakeCameraViewportActive\x12..MujocoMessage.MakeCameraViewportActiveRequest\x1a/.MujocoMessage.MakeCameraViewportActiveResponse\x12`\n\x0fSetMaterialInfo\x12%.MujocoMessage.SetMaterialInfoRequest\x1a&.MujocoMessage.SetMaterialInfoResponse\x12\x63\n\x10\x42\x65ginSaveMp4File\x12&.MujocoMessage.BeginSaveMp4FileRequest\x1a\'.MujocoMessage.BeginSaveMp4FileResponse\x12`\n\x0fStopSaveMp4File\x12%.MujocoMessage.StopSaveMp4FileRequest\x1a&.MujocoMessage.StopSaveMp4FileResponse\x12o\n\x14GetCurrentFrameIndex\x12*.MujocoMessage.GetCurrentFrameIndexRequest\x1a+.MujocoMessage.GetCurrentFrameIndexResponse\x12W\n\x0cGetTimeStamp\x12\".MujocoMessage.GetTimeStampRequest\x1a#.MujocoMessage.GetTimeStampResponse\x12\x66\n\x11GetCameraFramePNG\x12\'.MujocoMessage.GetCameraFramePNGRequest\x1a(.MujocoMessage.GetCameraFramePNGResponse\x12l\n\x13GetKeyPressedEvents\x12).MujocoMessage.GetKeyPressedEventsRequest\x1a*.MujocoMessage.GetKeyPressedEventsResponse\x12u\n\x16GetMouseMovementEvents\x12,.MujocoMessage.GetMouseMovementEventsRequest\x1a-.MujocoMessage.GetMouseMovementEventsResponse\x12\x84\x01\n\x1bGetBodyManipulationAnchored\x12\x31.MujocoMessage.GetBodyManipulationAnchoredRequest\x1a\x32.MujocoMessage.GetBodyManipulationAnchoredResponse\x12\x84\x01\n\x1bGetBodyManipulationMovement\x12\x31.MujocoMessage.GetBodyManipulationMovementRequest\x1a\x32.MujocoMessage.GetBodyManipulationMovementResponse\x12x\n\x17SetActorAnimParamNumber\x12-.MujocoMessage.SetActorAnimParamNumberRequest\x1a..MujocoMessage.SetActorAnimParamNumberResponse\x12r\n\x15SetActorAnimParamBool\x12+.MujocoMessage.SetActorAnimParamBoolRequest\x1a,.MujocoMessage.SetActorAnimParamBoolResponse\x12x\n\x17SetActorAnimParamString\x12-.MujocoMessage.SetActorAnimParamStringRequest\x1a..MujocoMessage.SetActorAnimParamStringResponse\x12]\n\x0eQueryModelInfo\x12$.MujocoMessage.QueryModelInfoRequest\x1a%.MujocoMessage.QueryModelInfoResponse\x12]\n\x0eQueryOptConfig\x12$.MujocoMessage.QueryOptConfigRequest\x1a%.MujocoMessage.QueryOptConfigResponse\x12]\n\x0eSetOptTimestep\x12$.MujocoMessage.SetOptTimestepRequest\x1a%.MujocoMessage.SetOptTimestepResponse\x12W\n\x0cSetOptConfig\x12\".MujocoMessage.SetOptConfigRequest\x1a#.MujocoMessage.SetOptConfigResponse\x12]\n\x0eQueryAllBodies\x12$.MujocoMessage.QueryAllBodiesRequest\x1a%.MujocoMessage.QueryAllBodiesResponse\x12]\n\x0eQueryAllJoints\x12$.MujocoMessage.QueryAllJointsRequest\x1a%.MujocoMessage.QueryAllJointsResponse\x12\x66\n\x11QueryAllActuators\x12\'.MujocoMessage.QueryAllActuatorsRequest\x1a(.MujocoMessage.QueryAllActuatorsResponse\x12Z\n\rQueryAllGeoms\x12#.MujocoMessage.QueryAllGeomsRequest\x1a$.MujocoMessage.QueryAllGeomsResponse\x12l\n\x13QueryAllMocapBodies\x12).MujocoMessage.QueryAllMocapBodiesRequest\x1a*.MujocoMessage.QueryAllMocapBodiesResponse\x12\x84\x01\n\x1bQueryAllEqualityConstraints\x12\x31.MujocoMessage.QueryAllEqualityConstraintsRequest\x1a\x32.MujocoMessage.QueryAllEqualityConstraintsResponse\x12Z\n\rQueryAllSites\x12#.MujocoMessage.QueryAllSitesRequest\x1a$.MujocoMessage.QueryAllSitesResponse\x12o\n\x14QueryAllQposQvelQacc\x12*.MujocoMessage.QueryAllQposQvelQaccRequest\x1a+.MujocoMessage.QueryAllQposQvelQaccResponse\x12W\n\x0cQueryContact\x12\".MujocoMessage.QueryContactRequest\x1a#.MujocoMessage.QueryContactResponse\x12i\n\x12QueryContactSimple\x12(.MujocoMessage.QueryContactSimpleRequest\x1a).MujocoMessage.QueryContactSimpleResponse\x12\x63\n\x10QueryQfrcInverse\x12&.MujocoMessage.QueryQfrcInverseRequest\x1a\'.MujocoMessage.QueryQfrcInverseResponse\x12l\n\x13QueryActuatorMoment\x12).MujocoMessage.QueryActuatorMomentRequest\x1a*.MujocoMessage.QueryActuatorMomentResponse\x12\x66\n\x11QueryQfrcActuator\x12\'.MujocoMessage.QueryQfrcActuatorRequest\x1a(.MujocoMessage.QueryQfrcActuatorResponse\x12W\n\x0cQueryCfrcExt\x12\".MujocoMessage.QueryCfrcExtRequest\x1a#.MujocoMessage.QueryCfrcExtResponse\x12i\n\x12QueryActuatorForce\x12(.MujocoMessage.QueryActuatorForceRequest\x1a).MujocoMessage.QueryActuatorForceResponse\x12l\n\x13QueryBodyVelocities\x12).MujocoMessage.QueryBodyVelocitiesRequest\x1a*.MujocoMessage.QueryBodyVelocitiesResponse\x12Z\n\rQueryQfrcBias\x12#.MujocoMessage.QueryQfrcBiasRequest\x1a$.MujocoMessage.QueryQfrcBiasResponse\x12`\n\x0fQuerySubtreeCom\x12%.MujocoMessage.QuerySubtreeComRequest\x1a&.MujocoMessage.QuerySubtreeComResponse\x12`\n\x0fQuerySensorData\x12%.MujocoMessage.QuerySensorDataRequest\x1a&.MujocoMessage.QuerySensorDataResponse\x12H\n\x07SetQpos\x12\x1d.MujocoMessage.SetQposRequest\x1a\x1e.MujocoMessage.SetQposResponse\x12H\n\x07SetQacc\x12\x1d.MujocoMessage.SetQaccRequest\x1a\x1e.MujocoMessage.SetQaccResponse\x12H\n\x07SetCtrl\x12\x1d.MujocoMessage.SetCtrlRequest\x1a\x1e.MujocoMessage.SetCtrlResponse\x12H\n\x07SetQvel\x12\x1d.MujocoMessage.SetQvelRequest\x1a\x1e.MujocoMessage.SetQvelResponse\x12`\n\x0fQueryJointNames\x12%.MujocoMessage.QueryJointNamesRequest\x1a&.MujocoMessage.QueryJointNamesResponse\x12\x63\n\x10QueryJointDofadr\x12&.MujocoMessage.QueryJointDofadrRequest\x1a\'.MujocoMessage.QueryJointDofadrResponse\x12]\n\x0eQueryChildBody\x12$.MujocoMessage.QueryChildBodyRequest\x1a%.MujocoMessage.QueryChildBodyResponse\x12Z\n\rQueryBodyQpos\x12#.MujocoMessage.QueryBodyQposRequest\x1a$.MujocoMessage.QueryBodyQposResponse\x12Z\n\rQueryBodyQvel\x12#.MujocoMessage.QueryBodyQvelRequest\x1a$.MujocoMessage.QueryBodyQvelResponse\x12]\n\x0eQueryJointQpos\x12$.MujocoMessage.QueryJointQposRequest\x1a%.MujocoMessage.QueryJointQposResponse\x12]\n\x0eQueryJointQvel\x12$.MujocoMessage.QueryJointQvelRequest\x1a%.MujocoMessage.QueryJointQvelResponse\x12\x81\x01\n\x1aQueryBodySubtreeMassByName\x12\x30.MujocoMessage.QueryBodySubtreeMassByNameRequest\x1a\x31.MujocoMessage.QueryBodySubtreeMassByNameResponse\x12\x63\n\x10QueryJointLimits\x12&.MujocoMessage.QueryJointLimitsRequest\x1a\'.MujocoMessage.QueryJointLimitsResponse\x12o\n\x14QueryActuatorGainPrm\x12*.MujocoMessage.QueryActuatorGainPrmRequest\x1a+.MujocoMessage.QueryActuatorGainPrmResponse\x12o\n\x14QueryActuatorBiasPrm\x12*.MujocoMessage.QueryActuatorBiasPrmRequest\x1a+.MujocoMessage.QueryActuatorBiasPrmResponse\x12i\n\x12QuerySitePosAndMat\x12(.MujocoMessage.QuerySitePosAndMatRequest\x1a).MujocoMessage.QuerySitePosAndMatResponse\x12W\n\x0cQuerySiteJac\x12\".MujocoMessage.QuerySiteJacRequest\x1a#.MujocoMessage.QuerySiteJacResponse\x12i\n\x12QueryBodyComPosMat\x12(.MujocoMessage.QueryBodyComPosMatRequest\x1a).MujocoMessage.QueryBodyComPosMatResponse\x12l\n\x13QueryBodyPosMatQuat\x12).MujocoMessage.QueryBodyPosMatQuatRequest\x1a*.MujocoMessage.QueryBodyPosMatQuatResponse\x12`\n\x0fQueryGeomPosMat\x12%.MujocoMessage.QueryGeomPosMatRequest\x1a&.MujocoMessage.QueryGeomPosMatResponse\x12\x66\n\x11QueryContactForce\x12\'.MujocoMessage.QueryContactForceRequest\x1a(.MujocoMessage.QueryContactForceResponse\x12\x66\n\x11QueryJointOffsets\x12\'.MujocoMessage.QueryJointOffsetsRequest\x1a(.MujocoMessage.QueryJointOffsetsResponse\x12`\n\x0fSetControlInput\x12%.MujocoMessage.SetControlInputRequest\x1a&.MujocoMessage.SetControlInputResponse\x12W\n\x0cSetJointQpos\x12\".MujocoMessage.SetJointQposRequest\x1a#.MujocoMessage.SetJointQposResponse\x12i\n\x12SetActuatorGainPrm\x12(.MujocoMessage.SetActuatorGainPrmRequest\x1a).MujocoMessage.SetActuatorGainPrmResponse\x12i\n\x12SetActuatorBiasPrm\x12(.MujocoMessage.SetActuatorBiasPrmRequest\x1a).MujocoMessage.SetActuatorBiasPrmResponse\x12o\n\x14QueryMocapPosAndQuat\x12*.MujocoMessage.QueryMocapPosAndQuatRequest\x1a+.MujocoMessage.QueryMocapPosAndQuatResponse\x12i\n\x12SetMocapPosAndQuat\x12(.MujocoMessage.SetMocapPosAndQuatRequest\x1a).MujocoMessage.SetMocapPosAndQuatResponse\x12~\n\x19UpdateEqualityConstraints\x12/.MujocoMessage.UpdateEqualityConstraintsRequest\x1a\x30.MujocoMessage.UpdateEqualityConstraintsResponse\x12`\n\x0fSetGeomFriction\x12%.MujocoMessage.SetGeomFrictionRequest\x1a&.MujocoMessage.SetGeomFrictionResponse\x12Q\n\nMJ_Forward\x12 .MujocoMessage.MJ_ForwardRequest\x1a!.MujocoMessage.MJ_ForwardResponse\x12Q\n\nMJ_Inverse\x12 .MujocoMessage.MJ_InverseRequest\x1a!.MujocoMessage.MJ_InverseResponse\x12H\n\x07MJ_Step\x12\x1d.MujocoMessage.MJ_StepRequest\x1a\x1e.MujocoMessage.MJ_StepResponse\x12{\n\x18\x43\x61lculateStateDifference\x12..MujocoMessage.CalculateStateDifferenceRequest\x1a/.MujocoMessage.CalculateStateDifferenceResponse\x12\x66\n\x13MJ_DifferentiatePos\x12&.MujocoMessage.DifferentiatePosRequest\x1a\'.MujocoMessage.DifferentiatePosResponse\x12[\n\x10MJD_TransitionFD\x12\".MujocoMessage.TransitionFDRequest\x1a#.MujocoMessage.TransitionFDResponse\x12]\n\x10MJ_JacSubtreeCom\x12#.MujocoMessage.JacSubtreeComRequest\x1a$.MujocoMessage.JacSubtreeComResponse\x12T\n\rMJ_JacBodyCom\x12 .MujocoMessage.JacBodyComRequest\x1a!.MujocoMessage.JacBodyComResponse\x12\x45\n\x06MJ_Jac\x12\x1c.MujocoMessage.MJ_JacRequest\x1a\x1d.MujocoMessage.MJ_JacResponse\x12i\n\x12\x43\x61lcFullMassMatrix\x12(.MujocoMessage.CalcFullMassMatrixRequest\x1a).MujocoMessage.CalcFullMassMatrixResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mjc_message_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GETTIMESTAMPRESPONSE_TIMESTAMPMAPENTRY']._loaded_options = None
  _globals['_GETTIMESTAMPRESPONSE_TIMESTAMPMAPENTRY']._serialized_options = b'8\001'
  _globals['_SIMULATIONSTATE']._serialized_start=18027
  _globals['_SIMULATIONSTATE']._serialized_end=18069
  _globals['_SETSIMULATIONSTATEREQUEST']._serialized_start=36
  _globals['_SETSIMULATIONSTATEREQUEST']._serialized_end=110
  _globals['_SETSIMULATIONSTATERESPONSE']._serialized_start=112
  _globals['_SETSIMULATIONSTATERESPONSE']._serialized_end=140
  _globals['_QUERYMODELINFOREQUEST']._serialized_start=142
  _globals['_QUERYMODELINFOREQUEST']._serialized_end=165
  _globals['_QUERYMODELINFORESPONSE']._serialized_start=168
  _globals['_QUERYMODELINFORESPONSE']._serialized_end=479
  _globals['_QUERYOPTCONFIGREQUEST']._serialized_start=481
  _globals['_QUERYOPTCONFIGREQUEST']._serialized_end=504
  _globals['_QUERYOPTCONFIGRESPONSE']._serialized_start=507
  _globals['_QUERYOPTCONFIGRESPONSE']._serialized_end=1111
  _globals['_SETOPTCONFIGREQUEST']._serialized_start=1114
  _globals['_SETOPTCONFIGREQUEST']._serialized_end=1715
  _globals['_SETOPTCONFIGRESPONSE']._serialized_start=1717
  _globals['_SETOPTCONFIGRESPONSE']._serialized_end=1739
  _globals['_QUERYALLBODIESREQUEST']._serialized_start=1741
  _globals['_QUERYALLBODIESREQUEST']._serialized_end=1764
  _globals['_QUERYALLBODIESRESPONSE']._serialized_start=1767
  _globals['_QUERYALLBODIESRESPONSE']._serialized_end=2289
  _globals['_QUERYALLBODIESRESPONSE_BODYINFO']._serialized_start=1861
  _globals['_QUERYALLBODIESRESPONSE_BODYINFO']._serialized_end=2289
  _globals['_QUERYALLJOINTSREQUEST']._serialized_start=2291
  _globals['_QUERYALLJOINTSREQUEST']._serialized_end=2314
  _globals['_QUERYALLJOINTSRESPONSE']._serialized_start=2317
  _globals['_QUERYALLJOINTSRESPONSE']._serialized_end=2714
  _globals['_QUERYALLJOINTSRESPONSE_JOINTINFO']._serialized_start=2413
  _globals['_QUERYALLJOINTSRESPONSE_JOINTINFO']._serialized_end=2714
  _globals['_QUERYJOINTNAMESREQUEST']._serialized_start=2716
  _globals['_QUERYJOINTNAMESREQUEST']._serialized_end=2740
  _globals['_QUERYJOINTNAMESRESPONSE']._serialized_start=2742
  _globals['_QUERYJOINTNAMESRESPONSE']._serialized_end=2787
  _globals['_QUERYJOINTDOFADRREQUEST']._serialized_start=2789
  _globals['_QUERYJOINTDOFADRREQUEST']._serialized_end=2834
  _globals['_QUERYJOINTDOFADRRESPONSE']._serialized_start=2836
  _globals['_QUERYJOINTDOFADRRESPONSE']._serialized_end=2884
  _globals['_QUERYCHILDBODYREQUEST']._serialized_start=2886
  _globals['_QUERYCHILDBODYREQUEST']._serialized_end=2933
  _globals['_QUERYCHILDBODYRESPONSE']._serialized_start=2935
  _globals['_QUERYCHILDBODYRESPONSE']._serialized_end=2985
  _globals['_QUERYBODYQPOSREQUEST']._serialized_start=2987
  _globals['_QUERYBODYQPOSREQUEST']._serialized_end=3031
  _globals['_JOINTQPOS']._serialized_start=3033
  _globals['_JOINTQPOS']._serialized_end=3077
  _globals['_BODYQPOS']._serialized_start=3079
  _globals['_BODYQPOS']._serialized_end=3149
  _globals['_QUERYBODYQPOSRESPONSE']._serialized_start=3151
  _globals['_QUERYBODYQPOSRESPONSE']._serialized_end=3221
  _globals['_QUERYBODYQVELREQUEST']._serialized_start=3223
  _globals['_QUERYBODYQVELREQUEST']._serialized_end=3267
  _globals['_JOINTQVEL']._serialized_start=3269
  _globals['_JOINTQVEL']._serialized_end=3313
  _globals['_BODYQVEL']._serialized_start=3315
  _globals['_BODYQVEL']._serialized_end=3385
  _globals['_QUERYBODYQVELRESPONSE']._serialized_start=3387
  _globals['_QUERYBODYQVELRESPONSE']._serialized_end=3457
  _globals['_SETCONTROLINPUTREQUEST']._serialized_start=3460
  _globals['_SETCONTROLINPUTREQUEST']._serialized_end=3620
  _globals['_SETCONTROLINPUTREQUEST_ACTUATORCONTROL']._serialized_start=3559
  _globals['_SETCONTROLINPUTREQUEST_ACTUATORCONTROL']._serialized_end=3620
  _globals['_SETCONTROLINPUTRESPONSE']._serialized_start=3623
  _globals['_SETCONTROLINPUTRESPONSE']._serialized_end=3777
  _globals['_SETCONTROLINPUTRESPONSE_ACTUATORRESULT']._serialized_start=3722
  _globals['_SETCONTROLINPUTRESPONSE_ACTUATORRESULT']._serialized_end=3777
  _globals['_LOADINITIALFRAMEREQUEST']._serialized_start=3779
  _globals['_LOADINITIALFRAMEREQUEST']._serialized_end=3804
  _globals['_LOADINITIALFRAMERESPONSE']._serialized_start=3806
  _globals['_LOADINITIALFRAMERESPONSE']._serialized_end=3832
  _globals['_LOADKEYFRAMEREQUEST']._serialized_start=3834
  _globals['_LOADKEYFRAMEREQUEST']._serialized_end=3877
  _globals['_LOADKEYFRAMERESPONSE']._serialized_start=3879
  _globals['_LOADKEYFRAMERESPONSE']._serialized_end=3918
  _globals['_QUERYALLACTUATORSREQUEST']._serialized_start=3920
  _globals['_QUERYALLACTUATORSREQUEST']._serialized_end=3946
  _globals['_QUERYALLACTUATORSRESPONSE']._serialized_start=3949
  _globals['_QUERYALLACTUATORSRESPONSE']._serialized_end=4748
  _globals['_QUERYALLACTUATORSRESPONSE_ACTUATORDATA']._serialized_start=4060
  _globals['_QUERYALLACTUATORSRESPONSE_ACTUATORDATA']._serialized_end=4748
  _globals['_QUERYALLGEOMSREQUEST']._serialized_start=4750
  _globals['_QUERYALLGEOMSREQUEST']._serialized_end=4772
  _globals['_QUERYALLGEOMSRESPONSE']._serialized_start=4775
  _globals['_QUERYALLGEOMSRESPONSE']._serialized_end=5318
  _globals['_QUERYALLGEOMSRESPONSE_GEOMDATA']._serialized_start=4872
  _globals['_QUERYALLGEOMSRESPONSE_GEOMDATA']._serialized_end=5318
  _globals['_QUERYJOINTQPOSREQUEST']._serialized_start=5320
  _globals['_QUERYJOINTQPOSREQUEST']._serialized_end=5366
  _globals['_QUERYJOINTQPOSRESPONSE']._serialized_start=5368
  _globals['_QUERYJOINTQPOSRESPONSE']._serialized_end=5441
  _globals['_QUERYJOINTQVELREQUEST']._serialized_start=5443
  _globals['_QUERYJOINTQVELREQUEST']._serialized_end=5489
  _globals['_QUERYJOINTQVELRESPONSE']._serialized_start=5491
  _globals['_QUERYJOINTQVELRESPONSE']._serialized_end=5564
  _globals['_CALCULATESTATEDIFFERENCEREQUEST']._serialized_start=5566
  _globals['_CALCULATESTATEDIFFERENCEREQUEST']._serialized_end=5640
  _globals['_STATEDIFFERENCE']._serialized_start=5642
  _globals['_STATEDIFFERENCE']._serialized_end=5704
  _globals['_CALCULATESTATEDIFFERENCERESPONSE']._serialized_start=5706
  _globals['_CALCULATESTATEDIFFERENCERESPONSE']._serialized_end=5795
  _globals['_DIFFERENTIATEPOSREQUEST']._serialized_start=5797
  _globals['_DIFFERENTIATEPOSREQUEST']._serialized_end=5857
  _globals['_DIFFERENTIATEPOSRESPONSE']._serialized_start=5859
  _globals['_DIFFERENTIATEPOSRESPONSE']._serialized_end=5897
  _globals['_TRANSITIONFDREQUEST']._serialized_start=5899
  _globals['_TRANSITIONFDREQUEST']._serialized_end=5958
  _globals['_TRANSITIONFDRESPONSE']._serialized_start=5960
  _globals['_TRANSITIONFDRESPONSE']._serialized_end=6028
  _globals['_JACSUBTREECOMREQUEST']._serialized_start=6030
  _globals['_JACSUBTREECOMREQUEST']._serialized_end=6070
  _globals['_JACSUBTREECOMRESPONSE']._serialized_start=6072
  _globals['_JACSUBTREECOMRESPONSE']._serialized_end=6111
  _globals['_JACBODYCOMREQUEST']._serialized_start=6113
  _globals['_JACBODYCOMREQUEST']._serialized_end=6150
  _globals['_JACBODYCOMRESPONSE']._serialized_start=6152
  _globals['_JACBODYCOMRESPONSE']._serialized_end=6200
  _globals['_QUERYALLQPOSQVELQACCREQUEST']._serialized_start=6202
  _globals['_QUERYALLQPOSQVELQACCREQUEST']._serialized_end=6231
  _globals['_QUERYALLQPOSQVELQACCRESPONSE']._serialized_start=6233
  _globals['_QUERYALLQPOSQVELQACCRESPONSE']._serialized_end=6305
  _globals['_QUERYCONTACTREQUEST']._serialized_start=6307
  _globals['_QUERYCONTACTREQUEST']._serialized_end=6328
  _globals['_QUERYCONTACTRESPONSE']._serialized_start=6331
  _globals['_QUERYCONTACTRESPONSE']._serialized_end=6735
  _globals['_QUERYCONTACTRESPONSE_CONTACTDATA']._serialized_start=6423
  _globals['_QUERYCONTACTRESPONSE_CONTACTDATA']._serialized_end=6735
  _globals['_QUERYCONTACTSIMPLEREQUEST']._serialized_start=6737
  _globals['_QUERYCONTACTSIMPLEREQUEST']._serialized_end=6764
  _globals['_QUERYCONTACTSIMPLERESPONSE']._serialized_start=6767
  _globals['_QUERYCONTACTSIMPLERESPONSE']._serialized_end=6926
  _globals['_QUERYCONTACTSIMPLERESPONSE_CONTACTDATA']._serialized_start=6870
  _globals['_QUERYCONTACTSIMPLERESPONSE_CONTACTDATA']._serialized_end=6926
  _globals['_QUERYQFRCINVERSEREQUEST']._serialized_start=6928
  _globals['_QUERYQFRCINVERSEREQUEST']._serialized_end=6953
  _globals['_QUERYQFRCINVERSERESPONSE']._serialized_start=6955
  _globals['_QUERYQFRCINVERSERESPONSE']._serialized_end=7003
  _globals['_QUERYACTUATORMOMENTREQUEST']._serialized_start=7005
  _globals['_QUERYACTUATORMOMENTREQUEST']._serialized_end=7033
  _globals['_QUERYACTUATORMOMENTRESPONSE']._serialized_start=7035
  _globals['_QUERYACTUATORMOMENTRESPONSE']._serialized_end=7113
  _globals['_QUERYQFRCACTUATORREQUEST']._serialized_start=7115
  _globals['_QUERYQFRCACTUATORREQUEST']._serialized_end=7141
  _globals['_QUERYQFRCACTUATORRESPONSE']._serialized_start=7143
  _globals['_QUERYQFRCACTUATORRESPONSE']._serialized_end=7193
  _globals['_SETQPOSREQUEST']._serialized_start=7195
  _globals['_SETQPOSREQUEST']._serialized_end=7225
  _globals['_SETQPOSRESPONSE']._serialized_start=7227
  _globals['_SETQPOSRESPONSE']._serialized_end=7244
  _globals['_SETQACCREQUEST']._serialized_start=7246
  _globals['_SETQACCREQUEST']._serialized_end=7276
  _globals['_SETQACCRESPONSE']._serialized_start=7278
  _globals['_SETQACCRESPONSE']._serialized_end=7295
  _globals['_QUERYBODYSUBTREEMASSBYNAMEREQUEST']._serialized_start=7297
  _globals['_QUERYBODYSUBTREEMASSBYNAMEREQUEST']._serialized_end=7351
  _globals['_QUERYBODYSUBTREEMASSBYNAMERESPONSE']._serialized_start=7353
  _globals['_QUERYBODYSUBTREEMASSBYNAMERESPONSE']._serialized_end=7415
  _globals['_SETOPTTIMESTEPREQUEST']._serialized_start=7417
  _globals['_SETOPTTIMESTEPREQUEST']._serialized_end=7458
  _globals['_SETOPTTIMESTEPRESPONSE']._serialized_start=7460
  _globals['_SETOPTTIMESTEPRESPONSE']._serialized_end=7484
  _globals['_SETCTRLREQUEST']._serialized_start=7486
  _globals['_SETCTRLREQUEST']._serialized_end=7516
  _globals['_SETCTRLRESPONSE']._serialized_start=7518
  _globals['_SETCTRLRESPONSE']._serialized_end=7535
  _globals['_MJ_FORWARDREQUEST']._serialized_start=7537
  _globals['_MJ_FORWARDREQUEST']._serialized_end=7556
  _globals['_MJ_FORWARDRESPONSE']._serialized_start=7558
  _globals['_MJ_FORWARDRESPONSE']._serialized_end=7578
  _globals['_MJ_INVERSEREQUEST']._serialized_start=7580
  _globals['_MJ_INVERSEREQUEST']._serialized_end=7599
  _globals['_MJ_INVERSERESPONSE']._serialized_start=7601
  _globals['_MJ_INVERSERESPONSE']._serialized_end=7621
  _globals['_MJ_STEPREQUEST']._serialized_start=7623
  _globals['_MJ_STEPREQUEST']._serialized_end=7654
  _globals['_MJ_STEPRESPONSE']._serialized_start=7656
  _globals['_MJ_STEPRESPONSE']._serialized_end=7673
  _globals['_SETQVELREQUEST']._serialized_start=7675
  _globals['_SETQVELREQUEST']._serialized_end=7705
  _globals['_SETQVELRESPONSE']._serialized_start=7707
  _globals['_SETQVELRESPONSE']._serialized_end=7724
  _globals['_QUERYCFRCEXTREQUEST']._serialized_start=7726
  _globals['_QUERYCFRCEXTREQUEST']._serialized_end=7767
  _globals['_QUERYCFRCEXTRESPONSE']._serialized_start=7769
  _globals['_QUERYCFRCEXTRESPONSE']._serialized_end=7860
  _globals['_BODYCFRCEXT']._serialized_start=7862
  _globals['_BODYCFRCEXT']._serialized_end=7912
  _globals['_SETJOINTQPOSREQUEST']._serialized_start=7914
  _globals['_SETJOINTQPOSREQUEST']._serialized_end=7986
  _globals['_SETJOINTQPOSRESPONSE']._serialized_start=7988
  _globals['_SETJOINTQPOSRESPONSE']._serialized_end=8027
  _globals['_QUERYACTUATORFORCEREQUEST']._serialized_start=8029
  _globals['_QUERYACTUATORFORCEREQUEST']._serialized_end=8056
  _globals['_QUERYACTUATORFORCERESPONSE']._serialized_start=8058
  _globals['_QUERYACTUATORFORCERESPONSE']._serialized_end=8110
  _globals['_QUERYJOINTLIMITSREQUEST']._serialized_start=8112
  _globals['_QUERYJOINTLIMITSREQUEST']._serialized_end=8158
  _globals['_JOINTLIMIT']._serialized_start=8160
  _globals['_JOINTLIMIT']._serialized_end=8249
  _globals['_QUERYJOINTLIMITSRESPONSE']._serialized_start=8251
  _globals['_QUERYJOINTLIMITSRESPONSE']._serialized_end=8326
  _globals['_QUERYBODYVELOCITIESREQUEST']._serialized_start=8328
  _globals['_QUERYBODYVELOCITIESREQUEST']._serialized_end=8376
  _globals['_QUERYBODYVELOCITIESRESPONSE']._serialized_start=8379
  _globals['_QUERYBODYVELOCITIESRESPONSE']._serialized_end=8576
  _globals['_QUERYBODYVELOCITIESRESPONSE_BODYVELOCITY']._serialized_start=8492
  _globals['_QUERYBODYVELOCITIESRESPONSE_BODYVELOCITY']._serialized_end=8576
  _globals['_QUERYACTUATORGAINPRMREQUEST']._serialized_start=8578
  _globals['_QUERYACTUATORGAINPRMREQUEST']._serialized_end=8631
  _globals['_ACTUATORGAINPRM']._serialized_start=8633
  _globals['_ACTUATORGAINPRM']._serialized_end=8691
  _globals['_QUERYACTUATORGAINPRMRESPONSE']._serialized_start=8693
  _globals['_QUERYACTUATORGAINPRMRESPONSE']._serialized_end=8778
  _globals['_SETACTUATORGAINPRMREQUEST']._serialized_start=8781
  _globals['_SETACTUATORGAINPRMREQUEST']._serialized_end=8959
  _globals['_SETACTUATORGAINPRMREQUEST_ACTUATORGAINPRMSET']._serialized_start=8898
  _globals['_SETACTUATORGAINPRMREQUEST_ACTUATORGAINPRMSET']._serialized_end=8959
  _globals['_SETACTUATORGAINPRMRESPONSE']._serialized_start=8961
  _globals['_SETACTUATORGAINPRMRESPONSE']._serialized_end=9006
  _globals['_QUERYACTUATORBIASPRMREQUEST']._serialized_start=9008
  _globals['_QUERYACTUATORBIASPRMREQUEST']._serialized_end=9061
  _globals['_ACTUATORBIASPRM']._serialized_start=9063
  _globals['_ACTUATORBIASPRM']._serialized_end=9121
  _globals['_QUERYACTUATORBIASPRMRESPONSE']._serialized_start=9123
  _globals['_QUERYACTUATORBIASPRMRESPONSE']._serialized_end=9208
  _globals['_SETACTUATORBIASPRMREQUEST']._serialized_start=9211
  _globals['_SETACTUATORBIASPRMREQUEST']._serialized_end=9389
  _globals['_SETACTUATORBIASPRMREQUEST_ACTUATORBIASPRMSET']._serialized_start=9328
  _globals['_SETACTUATORBIASPRMREQUEST_ACTUATORBIASPRMSET']._serialized_end=9389
  _globals['_SETACTUATORBIASPRMRESPONSE']._serialized_start=9391
  _globals['_SETACTUATORBIASPRMRESPONSE']._serialized_end=9436
  _globals['_QUERYALLMOCAPBODIESREQUEST']._serialized_start=9438
  _globals['_QUERYALLMOCAPBODIESREQUEST']._serialized_end=9466
  _globals['_QUERYALLMOCAPBODIESRESPONSE']._serialized_start=9469
  _globals['_QUERYALLMOCAPBODIESRESPONSE']._serialized_end=9643
  _globals['_QUERYALLMOCAPBODIESRESPONSE_MOCAPBODYINFO']._serialized_start=9580
  _globals['_QUERYALLMOCAPBODIESRESPONSE_MOCAPBODYINFO']._serialized_end=9643
  _globals['_QUERYMOCAPPOSANDQUATREQUEST']._serialized_start=9645
  _globals['_QUERYMOCAPPOSANDQUATREQUEST']._serialized_end=9700
  _globals['_QUERYMOCAPPOSANDQUATRESPONSE']._serialized_start=9703
  _globals['_QUERYMOCAPPOSANDQUATRESPONSE']._serialized_end=9886
  _globals['_QUERYMOCAPPOSANDQUATRESPONSE_MOCAPBODYINFO']._serialized_start=9819
  _globals['_QUERYMOCAPPOSANDQUATRESPONSE_MOCAPBODYINFO']._serialized_end=9886
  _globals['_SETMOCAPPOSANDQUATREQUEST']._serialized_start=9889
  _globals['_SETMOCAPPOSANDQUATREQUEST']._serialized_end=10066
  _globals['_SETMOCAPPOSANDQUATREQUEST_MOCAPBODYINFO']._serialized_start=9819
  _globals['_SETMOCAPPOSANDQUATREQUEST_MOCAPBODYINFO']._serialized_end=9886
  _globals['_SETMOCAPPOSANDQUATRESPONSE']._serialized_start=10068
  _globals['_SETMOCAPPOSANDQUATRESPONSE']._serialized_end=10113
  _globals['_QUERYALLEQUALITYCONSTRAINTSREQUEST']._serialized_start=10115
  _globals['_QUERYALLEQUALITYCONSTRAINTSREQUEST']._serialized_end=10151
  _globals['_QUERYALLEQUALITYCONSTRAINTSRESPONSE']._serialized_start=10154
  _globals['_QUERYALLEQUALITYCONSTRAINTSRESPONSE']._serialized_end=10436
  _globals['_QUERYALLEQUALITYCONSTRAINTSRESPONSE_EQUALITYCONSTRAINTINFO']._serialized_start=10299
  _globals['_QUERYALLEQUALITYCONSTRAINTSRESPONSE_EQUALITYCONSTRAINTINFO']._serialized_end=10436
  _globals['_QUERYSITEPOSANDMATREQUEST']._serialized_start=10438
  _globals['_QUERYSITEPOSANDMATREQUEST']._serialized_end=10485
  _globals['_QUERYSITEPOSANDMATRESPONSE']._serialized_start=10488
  _globals['_QUERYSITEPOSANDMATRESPONSE']._serialized_end=10663
  _globals['_QUERYSITEPOSANDMATRESPONSE_SITEPOSANDMAT']._serialized_start=10601
  _globals['_QUERYSITEPOSANDMATRESPONSE_SITEPOSANDMAT']._serialized_end=10663
  _globals['_QUERYSITEJACREQUEST']._serialized_start=10665
  _globals['_QUERYSITEJACREQUEST']._serialized_end=10706
  _globals['_QUERYSITEJACRESPONSE']._serialized_start=10709
  _globals['_QUERYSITEJACRESPONSE']._serialized_end=10853
  _globals['_QUERYSITEJACRESPONSE_SITEJAC']._serialized_start=10797
  _globals['_QUERYSITEJACRESPONSE_SITEJAC']._serialized_end=10853
  _globals['_UPDATEEQUALITYCONSTRAINTSREQUEST']._serialized_start=10856
  _globals['_UPDATEEQUALITYCONSTRAINTSREQUEST']._serialized_end=11068
  _globals['_UPDATEEQUALITYCONSTRAINTSREQUEST_EQUALITYCONSTRAINTSINFO']._serialized_start=10995
  _globals['_UPDATEEQUALITYCONSTRAINTSREQUEST_EQUALITYCONSTRAINTSINFO']._serialized_end=11068
  _globals['_UPDATEEQUALITYCONSTRAINTSRESPONSE']._serialized_start=11070
  _globals['_UPDATEEQUALITYCONSTRAINTSRESPONSE']._serialized_end=11122
  _globals['_QUERYBODYCOMPOSMATREQUEST']._serialized_start=11124
  _globals['_QUERYBODYCOMPOSMATREQUEST']._serialized_end=11175
  _globals['_QUERYBODYCOMPOSMATRESPONSE']._serialized_start=11178
  _globals['_QUERYBODYCOMPOSMATRESPONSE']._serialized_end=11364
  _globals['_QUERYBODYCOMPOSMATRESPONSE_BODYCOMPOSMAT']._serialized_start=11296
  _globals['_QUERYBODYCOMPOSMATRESPONSE_BODYCOMPOSMAT']._serialized_end=11364
  _globals['_QUERYBODYPOSMATQUATREQUEST']._serialized_start=11366
  _globals['_QUERYBODYPOSMATQUATREQUEST']._serialized_end=11418
  _globals['_QUERYBODYPOSMATQUATRESPONSE']._serialized_start=11421
  _globals['_QUERYBODYPOSMATQUATRESPONSE']._serialized_end=11618
  _globals['_QUERYBODYPOSMATQUATRESPONSE_BODYPOSMATQUAT']._serialized_start=11543
  _globals['_QUERYBODYPOSMATQUATRESPONSE_BODYPOSMATQUAT']._serialized_end=11618
  _globals['_QUERYGEOMPOSMATREQUEST']._serialized_start=11620
  _globals['_QUERYGEOMPOSMATREQUEST']._serialized_end=11668
  _globals['_QUERYGEOMPOSMATRESPONSE']._serialized_start=11671
  _globals['_QUERYGEOMPOSMATRESPONSE']._serialized_end=11833
  _globals['_QUERYGEOMPOSMATRESPONSE_GEOMPOSMAT']._serialized_start=11776
  _globals['_QUERYGEOMPOSMATRESPONSE_GEOMPOSMAT']._serialized_end=11833
  _globals['_QUERYCONTACTFORCEREQUEST']._serialized_start=11835
  _globals['_QUERYCONTACTFORCEREQUEST']._serialized_end=11882
  _globals['_QUERYCONTACTFORCERESPONSE']._serialized_start=11885
  _globals['_QUERYCONTACTFORCERESPONSE']._serialized_end=12035
  _globals['_QUERYCONTACTFORCERESPONSE_CONTACTFORCE']._serialized_start=11993
  _globals['_QUERYCONTACTFORCERESPONSE_CONTACTFORCE']._serialized_end=12035
  _globals['_MJ_JACREQUEST']._serialized_start=12038
  _globals['_MJ_JACREQUEST']._serialized_end=12204
  _globals['_MJ_JACREQUEST_BODYPOINT']._serialized_start=12164
  _globals['_MJ_JACREQUEST_BODYPOINT']._serialized_end=12204
  _globals['_MJ_JACRESPONSE']._serialized_start=12206
  _globals['_MJ_JACRESPONSE']._serialized_end=12325
  _globals['_MJ_JACRESPONSE_JACRESULT']._serialized_start=12286
  _globals['_MJ_JACRESPONSE_JACRESULT']._serialized_end=12325
  _globals['_CALCFULLMASSMATRIXREQUEST']._serialized_start=12327
  _globals['_CALCFULLMASSMATRIXREQUEST']._serialized_end=12354
  _globals['_CALCFULLMASSMATRIXRESPONSE']._serialized_start=12356
  _globals['_CALCFULLMASSMATRIXRESPONSE']._serialized_end=12410
  _globals['_QUERYQFRCBIASREQUEST']._serialized_start=12412
  _globals['_QUERYQFRCBIASREQUEST']._serialized_end=12434
  _globals['_QUERYQFRCBIASRESPONSE']._serialized_start=12436
  _globals['_QUERYQFRCBIASRESPONSE']._serialized_end=12478
  _globals['_QUERYSUBTREECOMREQUEST']._serialized_start=12480
  _globals['_QUERYSUBTREECOMREQUEST']._serialized_end=12528
  _globals['_QUERYSUBTREECOMRESPONSE']._serialized_start=12530
  _globals['_QUERYSUBTREECOMRESPONSE']._serialized_end=12600
  _globals['_SETGEOMFRICTIONREQUEST']._serialized_start=12603
  _globals['_SETGEOMFRICTIONREQUEST']._serialized_end=12750
  _globals['_SETGEOMFRICTIONREQUEST_FRICTION']._serialized_start=12724
  _globals['_SETGEOMFRICTIONREQUEST_FRICTION']._serialized_end=12750
  _globals['_SETGEOMFRICTIONRESPONSE']._serialized_start=12752
  _globals['_SETGEOMFRICTIONRESPONSE']._serialized_end=12799
  _globals['_QUERYSENSORDATAREQUEST']._serialized_start=12801
  _globals['_QUERYSENSORDATAREQUEST']._serialized_end=12847
  _globals['_QUERYSENSORDATARESPONSE']._serialized_start=12850
  _globals['_QUERYSENSORDATARESPONSE']._serialized_end=13030
  _globals['_QUERYSENSORDATARESPONSE_SENSORDATA']._serialized_start=12954
  _globals['_QUERYSENSORDATARESPONSE_SENSORDATA']._serialized_end=13030
  _globals['_QUERYJOINTOFFSETSREQUEST']._serialized_start=13032
  _globals['_QUERYJOINTOFFSETSREQUEST']._serialized_end=13079
  _globals['_QUERYJOINTOFFSETSRESPONSE']._serialized_start=13082
  _globals['_QUERYJOINTOFFSETSRESPONSE']._serialized_end=13284
  _globals['_QUERYJOINTOFFSETSRESPONSE_JOINTOFFSET']._serialized_start=13188
  _globals['_QUERYJOINTOFFSETSRESPONSE_JOINTOFFSET']._serialized_end=13284
  _globals['_QUERYALLSITESREQUEST']._serialized_start=13286
  _globals['_QUERYALLSITESREQUEST']._serialized_end=13308
  _globals['_QUERYALLSITESRESPONSE']._serialized_start=13311
  _globals['_QUERYALLSITESRESPONSE']._serialized_end=13544
  _globals['_QUERYALLSITESRESPONSE_SITEINFO']._serialized_start=13403
  _globals['_QUERYALLSITESRESPONSE_SITEINFO']._serialized_end=13544
  _globals['_UPDATELOCALENVREQUEST']._serialized_start=13546
  _globals['_UPDATELOCALENVREQUEST']._serialized_end=13597
  _globals['_UPDATELOCALENVRESPONSE']._serialized_start=13600
  _globals['_UPDATELOCALENVRESPONSE']._serialized_end=13754
  _globals['_UPDATELOCALENVRESPONSE_OVERRIDECTRLINFO']._serialized_start=13706
  _globals['_UPDATELOCALENVRESPONSE_OVERRIDECTRLINFO']._serialized_end=13754
  _globals['_LOADLOCALENVREQUEST']._serialized_start=13757
  _globals['_LOADLOCALENVREQUEST']._serialized_end=13892
  _globals['_LOADLOCALENVREQUEST_REQTYPE']._serialized_start=13842
  _globals['_LOADLOCALENVREQUEST_REQTYPE']._serialized_end=13892
  _globals['_LOADLOCALENVRESPONSE']._serialized_start=13895
  _globals['_LOADLOCALENVRESPONSE']._serialized_end=14074
  _globals['_LOADLOCALENVRESPONSE_STATUS']._serialized_start=14042
  _globals['_LOADLOCALENVRESPONSE_STATUS']._serialized_end=14074
  _globals['_LOADCONTENTFILEREQUEST']._serialized_start=14076
  _globals['_LOADCONTENTFILEREQUEST']._serialized_end=14137
  _globals['_LOADCONTENTFILERESPONSE']._serialized_start=14140
  _globals['_LOADCONTENTFILERESPONSE']._serialized_end=14302
  _globals['_LOADCONTENTFILERESPONSE_STATUS']._serialized_start=14042
  _globals['_LOADCONTENTFILERESPONSE_STATUS']._serialized_end=14074
  _globals['_ADDACTORREQUEST']._serialized_start=14304
  _globals['_ADDACTORREQUEST']._serialized_end=14401
  _globals['_ADDACTORRESPONSE']._serialized_start=14404
  _globals['_ADDACTORRESPONSE']._serialized_end=14535
  _globals['_ADDACTORRESPONSE_STATUS']._serialized_start=14042
  _globals['_ADDACTORRESPONSE_STATUS']._serialized_end=14074
  _globals['_PUBLISHSCENEREQUEST']._serialized_start=14537
  _globals['_PUBLISHSCENEREQUEST']._serialized_end=14558
  _globals['_PUBLISHSCENERESPONSE']._serialized_start=14561
  _globals['_PUBLISHSCENERESPONSE']._serialized_end=14700
  _globals['_PUBLISHSCENERESPONSE_STATUS']._serialized_start=14042
  _globals['_PUBLISHSCENERESPONSE_STATUS']._serialized_end=14074
  _globals['_SETLIGHTINFOREQUEST']._serialized_start=14702
  _globals['_SETLIGHTINFOREQUEST']._serialized_end=14789
  _globals['_SETLIGHTINFORESPONSE']._serialized_start=14792
  _globals['_SETLIGHTINFORESPONSE']._serialized_end=14931
  _globals['_SETLIGHTINFORESPONSE_STATUS']._serialized_start=14042
  _globals['_SETLIGHTINFORESPONSE_STATUS']._serialized_end=14074
  _globals['_SETCAMERASENSORINFOREQUEST']._serialized_start=14934
  _globals['_SETCAMERASENSORINFOREQUEST']._serialized_end=15066
  _globals['_SETCAMERASENSORINFORESPONSE']._serialized_start=15069
  _globals['_SETCAMERASENSORINFORESPONSE']._serialized_end=15222
  _globals['_SETCAMERASENSORINFORESPONSE_STATUS']._serialized_start=14042
  _globals['_SETCAMERASENSORINFORESPONSE_STATUS']._serialized_end=14074
  _globals['_BEGINSAVEMP4FILEREQUEST']._serialized_start=15225
  _globals['_BEGINSAVEMP4FILEREQUEST']._serialized_end=15379
  _globals['_BEGINSAVEMP4FILEREQUEST_CAPTUREMODE']._serialized_start=15345
  _globals['_BEGINSAVEMP4FILEREQUEST_CAPTUREMODE']._serialized_end=15379
  _globals['_BEGINSAVEMP4FILERESPONSE']._serialized_start=15382
  _globals['_BEGINSAVEMP4FILERESPONSE']._serialized_end=15529
  _globals['_BEGINSAVEMP4FILERESPONSE_STATUS']._serialized_start=14042
  _globals['_BEGINSAVEMP4FILERESPONSE_STATUS']._serialized_end=14074
  _globals['_STOPSAVEMP4FILEREQUEST']._serialized_start=15531
  _globals['_STOPSAVEMP4FILEREQUEST']._serialized_end=15652
  _globals['_STOPSAVEMP4FILEREQUEST_STATUS']._serialized_start=15619
  _globals['_STOPSAVEMP4FILEREQUEST_STATUS']._serialized_end=15652
  _globals['_STOPSAVEMP4FILERESPONSE']._serialized_start=15654
  _globals['_STOPSAVEMP4FILERESPONSE']._serialized_end=15679
  _globals['_GETCURRENTFRAMEINDEXREQUEST']._serialized_start=15681
  _globals['_GETCURRENTFRAMEINDEXREQUEST']._serialized_end=15710
  _globals['_GETCURRENTFRAMEINDEXRESPONSE']._serialized_start=15712
  _globals['_GETCURRENTFRAMEINDEXRESPONSE']._serialized_end=15788
  _globals['_GETTIMESTAMPREQUEST']._serialized_start=15790
  _globals['_GETTIMESTAMPREQUEST']._serialized_end=15837
  _globals['_TIMESTAMPLIST']._serialized_start=15839
  _globals['_TIMESTAMPLIST']._serialized_end=15875
  _globals['_GETTIMESTAMPRESPONSE']._serialized_start=15878
  _globals['_GETTIMESTAMPRESPONSE']._serialized_end=16085
  _globals['_GETTIMESTAMPRESPONSE_TIMESTAMPMAPENTRY']._serialized_start=16004
  _globals['_GETTIMESTAMPRESPONSE_TIMESTAMPMAPENTRY']._serialized_end=16085
  _globals['_GETCAMERAFRAMEPNGREQUEST']._serialized_start=16087
  _globals['_GETCAMERAFRAMEPNGREQUEST']._serialized_end=16133
  _globals['_NAMEANDTRANSFORM']._serialized_start=16135
  _globals['_NAMEANDTRANSFORM']._serialized_end=16194
  _globals['_GETCAMERAFRAMEPNGRESPONSE']._serialized_start=16196
  _globals['_GETCAMERAFRAMEPNGRESPONSE']._serialized_end=16280
  _globals['_MAKECAMERAVIEWPORTACTIVEREQUEST']._serialized_start=16282
  _globals['_MAKECAMERAVIEWPORTACTIVEREQUEST']._serialized_end=16356
  _globals['_MAKECAMERAVIEWPORTACTIVERESPONSE']._serialized_start=16359
  _globals['_MAKECAMERAVIEWPORTACTIVERESPONSE']._serialized_end=16522
  _globals['_MAKECAMERAVIEWPORTACTIVERESPONSE_STATUS']._serialized_start=14042
  _globals['_MAKECAMERAVIEWPORTACTIVERESPONSE_STATUS']._serialized_end=14074
  _globals['_SETMATERIALINFOREQUEST']._serialized_start=16524
  _globals['_SETMATERIALINFOREQUEST']._serialized_end=16588
  _globals['_SETMATERIALINFORESPONSE']._serialized_start=16591
  _globals['_SETMATERIALINFORESPONSE']._serialized_end=16736
  _globals['_SETMATERIALINFORESPONSE_STATUS']._serialized_start=14042
  _globals['_SETMATERIALINFORESPONSE_STATUS']._serialized_end=14074
  _globals['_GETKEYPRESSEDEVENTSREQUEST']._serialized_start=16738
  _globals['_GETKEYPRESSEDEVENTSREQUEST']._serialized_end=16766
  _globals['_GETKEYPRESSEDEVENTSRESPONSE']._serialized_start=16768
  _globals['_GETKEYPRESSEDEVENTSRESPONSE']._serialized_end=16813
  _globals['_GETMOUSEMOVEMENTEVENTSREQUEST']._serialized_start=16815
  _globals['_GETMOUSEMOVEMENTEVENTSREQUEST']._serialized_end=16846
  _globals['_GETMOUSEMOVEMENTEVENTSRESPONSE']._serialized_start=16848
  _globals['_GETMOUSEMOVEMENTEVENTSRESPONSE']._serialized_end=16931
  _globals['_GETBODYMANIPULATIONANCHOREDREQUEST']._serialized_start=16933
  _globals['_GETBODYMANIPULATIONANCHOREDREQUEST']._serialized_end=16969
  _globals['_GETBODYMANIPULATIONANCHOREDRESPONSE']._serialized_start=16972
  _globals['_GETBODYMANIPULATIONANCHOREDRESPONSE']._serialized_end=17156
  _globals['_GETBODYMANIPULATIONANCHOREDRESPONSE_ANCHORTYPE']._serialized_start=17114
  _globals['_GETBODYMANIPULATIONANCHOREDRESPONSE_ANCHORTYPE']._serialized_end=17156
  _globals['_GETBODYMANIPULATIONMOVEMENTREQUEST']._serialized_start=17158
  _globals['_GETBODYMANIPULATIONMOVEMENTREQUEST']._serialized_end=17194
  _globals['_GETBODYMANIPULATIONMOVEMENTRESPONSE']._serialized_start=17196
  _globals['_GETBODYMANIPULATIONMOVEMENTRESPONSE']._serialized_end=17272
  _globals['_SETACTORANIMPARAMNUMBERREQUEST']._serialized_start=17274
  _globals['_SETACTORANIMPARAMNUMBERREQUEST']._serialized_end=17361
  _globals['_SETACTORANIMPARAMNUMBERRESPONSE']._serialized_start=17364
  _globals['_SETACTORANIMPARAMNUMBERRESPONSE']._serialized_end=17525
  _globals['_SETACTORANIMPARAMNUMBERRESPONSE_STATUS']._serialized_start=14042
  _globals['_SETACTORANIMPARAMNUMBERRESPONSE_STATUS']._serialized_end=14074
  _globals['_SETACTORANIMPARAMBOOLREQUEST']._serialized_start=17527
  _globals['_SETACTORANIMPARAMBOOLREQUEST']._serialized_end=17612
  _globals['_SETACTORANIMPARAMBOOLRESPONSE']._serialized_start=17615
  _globals['_SETACTORANIMPARAMBOOLRESPONSE']._serialized_end=17772
  _globals['_SETACTORANIMPARAMBOOLRESPONSE_STATUS']._serialized_start=14042
  _globals['_SETACTORANIMPARAMBOOLRESPONSE_STATUS']._serialized_end=14074
  _globals['_SETACTORANIMPARAMSTRINGREQUEST']._serialized_start=17774
  _globals['_SETACTORANIMPARAMSTRINGREQUEST']._serialized_end=17861
  _globals['_SETACTORANIMPARAMSTRINGRESPONSE']._serialized_start=17864
  _globals['_SETACTORANIMPARAMSTRINGRESPONSE']._serialized_end=18025
  _globals['_SETACTORANIMPARAMSTRINGRESPONSE_STATUS']._serialized_start=14042
  _globals['_SETACTORANIMPARAMSTRINGRESPONSE_STATUS']._serialized_end=14074
  _globals['_GRPCSERVICE']._serialized_start=18072
  _globals['_GRPCSERVICE']._serialized_end=26841
# @@protoc_insertion_point(module_scope)
